# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import copy
import time
import warnings
import shutil
import tempfile

import os
import numpy as np

import tensorflow as tf
import tensorflow.keras.optimizers as tfko
from tensorflow.keras.layers import Dense

from leap_net.proxy.baseProxy import BaseProxy


class BaseNNProxy(BaseProxy):
    """
    This class serves as base class for all proxy made from neural network with tensorflow

    We recommend to  inherit from this class if you are to code a neural network based proxy (using tensorflow).
    In particular the :func:`BaseNNProxy.train` function is not trivial to implement correctly.


    Attributes
    ----------
    train_iter: ``int``
        The current number of training iteration

    train_batch_size: ``int``
        An integer > 0. It represents the training batch size of the proxy. When the proxy is trained, it tells
        how many "state" are feed at once.

    _lr: ``float``
        The learning rate (discarded when the proxy do not need to be learned)

    _layer_fun: ``tensorflow.keras.layers``
        The "function" representing each layers. Be careful, this is not serialized and so if you want to
        save / reload the model, you need to specify the same function manually.

    _layer_act: ``str``
        The activation function of each layers. Should be a string.

    _optimizer_model:
        Represents the optimizer of the neural network

    _schedule_lr_model:
        internal, do not use
    """
    def __init__(self,
                 name,
                 lr=1e-4,
                 train_batch_size=32,
                 max_row_training_set=int(1e5),
                 eval_batch_size=1024,
                 layer=Dense,
                 layer_act=None,
                 attr_x=("prod_p", "prod_v", "load_p", "load_q", "topo_vect"),  # input that will be given to the proxy
                 attr_y=("a_or", "a_ex", "p_or", "p_ex", "q_or", "q_ex", "prod_q", "load_v", "v_or", "v_ex"),  # output that we want the proxy to predict
                 ):
        BaseProxy.__init__(self,
                           name=name,
                           max_row_training_set=max_row_training_set,
                           eval_batch_size=eval_batch_size,
                           attr_x=attr_x,
                           attr_y=attr_y
                           )

        self._layer_fun = layer
        self._layer_act = layer_act

        self._lr = lr

        self.train_iter = 0  # number of training iteration
        # model optimizer
        self._schedule_lr_model = None
        self._optimizer_model = None

        self.train_batch_size = train_batch_size
        if self.max_row_training_set < self.train_batch_size:
            raise RuntimeError(f"You cannot use a batch size of {self.train_batch_size} with a dataset counting at"
                               f" most {self.max_row_training_set} rows. "
                               "Please either increase \"max_row_training_set\" or decrease \"batch_size\""
                               "(hint: batch_size>=max_row_training_set).")

        self.__need_save_graph = True  # save the tensorflow computation graph

    def _make_predictions(self, data, training=False):
        """
        Make a prediction with the proxy on a new grid state.

        It's analogous to the `train_model` but instead of training it gives the prediction of the neural network.

        It's called with:

        .. code-block:: python

            data = self._extract_data([last_index])
            tmp = self._make_predictions(data)
            res = self._post_process(tmp)

        This function can be overridden (for example if your proxy does not use tensorflow)

        It's not part of the public API that is used outside of your proxy (private method).

        Parameters
        ----------
        data:
            The input data on which the model will make the predictions (list of numpy arrays)
        training: ``bool``
            Whether this method is used to make prediction during training, or not

        Returns
        -------
        res:
            A list of numpy array that will be post process afterwards. In particular
             this list must count as many elements as there
            are elements in `attr_y`.
        """
        return self._model(data, training=training)

    def load_metadata(self, dict_):
        """
        this function is used when loading the proxy to restore the meta data

        This function may be overridden but in that case we recommend to call the method of the super class (like
        this is done here when calling `super().load_metadata(dict_)` )

        Notes
        -----
        This function is expected to modify the instance on which it is called (*ie* `self`)

        Parameters
        ----------
        dict_: ``dict``
            The dictionary of parameter that is used to initialize this instance.

        """

        self.attr_x = tuple([str(el) for el in dict_["attr_x"]])
        self.attr_y = tuple([str(el) for el in dict_["attr_y"]])

        self._sz_x = [int(el) for el in dict_["_sz_x"]]
        self._sz_y = [int(el) for el in dict_["_sz_y"]]

        self._time_train = float(dict_["_time_train"])
        self._time_predict = float(dict_["_time_predict"])

        self._init_database_shapes()
        super().load_metadata(dict_)

    def get_metadata(self):
        """
        should return a dictionary containing all the metadata of this class in a format that is compatible
        with json serialization.

        This function may be overridden but in that case we recommend to call the method of the super class (like
        is done here)

        Notes
        -----
        We assume that the metadata are "json serializable".

        Returns
        --------
        res: ``dict``
            A dictionary containing the necessary information to initialize the class as when it was saved.

        """
        res = super().get_metadata()
        if self._layer_act is not None:
            res["_layer_act"] = str(self._layer_act)
        else:
            # i don't store anything if it's None
            pass

        return res

    def save_data(self, path, ext=".h5"):
        """
        Save extra information that might be required by the model. For example this saves the weights of
        some neural network if needed.

        This function should be overridden

        Notes
        -----
        We suppose that there is a "." preceding the extension. So ext=".h5" is valid, but not ext="h5"

        Parameters
        ----------
        path: ``str``
            The path at which the model will be saved
        ext: ``str``
            The extension used to save the model (for example ".h5" should output a file named xxx.h5)

        """
        _if_no_model = os.path.join(path, ".nomodel")
        if self._model is not None:
            # save the weights
            self._model.save(os.path.join(path, f"weights{ext}"))
            # remove the ".nomodel" if it exists
            if os.path.exists(_if_no_model):
               os.remove(_if_no_model)
        else:
            with open(_if_no_model, mode="w", encoding="utf-8") as f:
                f.write("EMPTY NEURAL NETWORK, IS THIS NORMAL ?\n")

    def load_data(self, path, ext=".h5"):
        """
        You need to override this function if the proxy, in order to be functional need to load some data
        that are not loaded by the `load_metadata` function.

        This function may be overridden but in that case we recommend to call the method of the super class

        load the weight of the neural network
        path is the full path (including file name and extension).

        This function can be overridden (for example if your proxy does not use tensorflow or is made of multiple
        submodule that need to be saved independently)

        This function is used when loading back your proxy.

        Notes
        -----
        This function is only called when the metadata (number of layer, size of each layer etc.)
        have been properly restored (so it supposes load_metadata has been successfully called)

        We suppose that there is a "." preceding the extension. So ext=".h5" is valid, but not ext="h5"

        We recommend that the file at the "path" is not directly read.
        Rather, it is better to first copied it into a temporary directory, and then
        read it. This is to avoid any data corruption when an instance of the model is reading and another
        is writing to the same file.

        """
        if os.path.exists(os.path.join(path, ".nomodel")):
            # no weights were present
            if self._model is not None:
                raise RuntimeError("The _model member is initialized, yet no weights were previously saved. "
                                   "Thus i cannot restore it.")
            # otherwise i do nothing: model un initialized and no weights stored is perfectly fine
        else:
            with tempfile.TemporaryDirectory() as path_tmp:
                nm_file = f"weights{ext}"
                nm_tmp = os.path.join(path_tmp, nm_file)
                # copy the weights into this file
                shutil.copy(os.path.join(path, nm_file), nm_tmp)
                # load this copy (make sure the proper file is not corrupted even if the loading fails)
                self._model.load_weights(nm_tmp)

    def _train_model(self, data):
        """
        perform the training step. For model coded in tensorflow in a regular supervised learning
        setting it can be summarize as follow:

        .. code-block:: python

            self._model.train_on_batch(*data)

        This function is called with something like:

        .. code-block:: python

            data = self._extract_data(indx_train)
            loss = self.train_model(data)

        This function can be overridden (for example if your proxy does not use tensorflow)

        It's not part of the public API that is used outside of your proxy (private method).

        Parameters
        ----------
        data:
            A tuple of np array that can be used for training

        Returns
        -------
        The loss of the batch

        """
        losses = self._model.train_on_batch(*data)
        return losses

    def _make_optimiser(self):
        """
        helper function to create the proper optimizer (Adam) with the learning rates and its decay
        parameters.

        This function can be overridden (for example if you don't use tensorflow).

        It's not part of the public API that is used outside of your proxy (private method).

        """
        # schedule = tfko.schedules.InverseTimeDecay(self._lr, self._lr_decay_steps, self._lr_decay_rate)
        return None, tfko.Adam(learning_rate=self._lr)

    def save_tensorboard(self, tf_writer, training_iter, batch_losses):
        """
        save extra information to tensorboard

        In this case i save all the losses for all individual output
        """
        for output_nm, loss in zip(self.attr_y, batch_losses):
            tf.summary.scalar(f"{output_nm}", loss, training_iter,
                              description=f"MSE for {output_nm}")

    #######################################################
    ## We don't recommend to change anything bellow this ##
    #######################################################
    def train(self, tf_writer=None):
        """
        Train the proxy (if tf_writer is not None, it is expected that the proxy save the computation graph

        We don't recommend to override this function.

        Parameters
        ----------
        tf_writer

        Returns
        -------
        None if the proxy has not been trained at this iteration, or the losses
        """
        if self._global_iter % self.train_batch_size != 0:
            return None

        if self._is_db_full():
            tmp_max = self.max_row_training_set
        else:
            tmp_max = self.last_id

        # TODO seed
        # TODO mode "i keep only the last of the dataset
        indx_train = np.random.choice(np.arange(tmp_max),
                                      size=self.train_batch_size,
                                      replace=False)

        if self.DEBUG:
            indx_train = np.arange(self.train_batch_size)

        data = self._extract_data(indx_train)

        # for el in data[1]: print(np.mean(el))
        if tf_writer is not None and self.__need_save_graph:
            tf.summary.trace_on()

        beg_ = time.time()
        batch_losses = self._train_model(data)
        self._time_train += time.time() - beg_
        if tf_writer is not None and self.__need_save_graph:
            with tf_writer.as_default():
                tf.summary.trace_export("model-graph", 0)
            self.__need_save_graph = False
            tf.summary.trace_off()
        return batch_losses

    def _get_mean(self, obss, attr_nm):
        """
        For the scaler, compute the mean that will be used to scale the data

        This function can be overridden (for example if you want more control on how to scale the data)

        obss is a list of observation
        """
        add_, mul = self._get_adds_mults_from_name(obss, attr_nm)
        return add_

    def _get_sd(self, obss, attr_nm):
        """
        For the scaler, compute the mean that will be used to scale the data

        This function can be overridden (for example if you want more control on how to scale the data)

        obss is a list of observation
        """
        add_, mul_ = self._get_adds_mults_from_name(obss, attr_nm)
        return mul_



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import grid2op
import numpy as np
import unittest
import warnings
from leap_net.proxy import ProxyLeapNet
import tempfile


class Test(unittest.TestCase):
    def setUp(self) -> None:
        self.proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                  topo_vect_to_tau="all")

        # valid only for this environment
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            self.env = grid2op.make("l2rpn_case14_sandbox", test=True)
        # the number of elements per substations are:
        # [3, 6, 4, 6, 5, 7, 3, 2, 5, 3, 3, 3, 4, 3]
        self.obs = self.env.reset()
        self.proxy.init([self.obs])

    def test_load_save(self):
        # test for "all"
        with tempfile.TemporaryDirectory() as path:
            self.proxy.save_data(path)
            proxy_loaded = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                        topo_vect_to_tau="all")
            proxy_loaded.load_data(path)
            self._aux_test_tau_from_topo_vect_all(proxy_loaded)

        # test the different load / save mechanics for the different encodings"
        with tempfile.TemporaryDirectory() as path:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="raw")
            proxy.init([self.obs])
            proxy.save_data(path)
            proxy_loaded = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                        topo_vect_to_tau="raw")
            proxy_loaded.load_data(path)
            self._aux_test_tau_default(proxy_loaded)

        with tempfile.TemporaryDirectory() as path:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="given_list",
                                 kwargs_tau=[(0, (2, 1, 1)), (0, (1, 2, 1)), (1, (2, 1, 1, 1, 1, 1)),
                                             (12, (2, 1, 1, 2)), (13, (2, 1, 2)), (13, (1, 2, 2))]
                                 )
            proxy.init([self.obs])
            proxy.save_data(path)
            proxy_loaded = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                        topo_vect_to_tau="given_list")
            proxy_loaded.load_data(path)
            self._aux_test_tau_from_list_topo(proxy_loaded)

        with tempfile.TemporaryDirectory() as path:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="online_list",
                                 kwargs_tau=7
                                 )
            proxy.init([self.obs])
            proxy.save_data(path)
            proxy_loaded = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                        topo_vect_to_tau="online_list",)
            proxy_loaded.load_data(path)
            self._aux_test_tau_from_online_topo(proxy_loaded)

    def test_tau_default(self):
        self._aux_test_tau_default()

    def _aux_test_tau_default(self, proxy=None):
        if proxy is None:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="raw")
            proxy.init([self.obs])

        # valid only for this environment
        env = self.env
        # the number of elements per substations are:
        # [3, 6, 4, 6, 5, 7, 3, 2, 5, 3, 3, 3, 4, 3]
        obs = env.reset()
        proxy.init([obs])

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[0] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (1, 2, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[1] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 1, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[3] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(12, (2, 2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 4
        assert np.all(res[[50, 51, 52, 53]] == 1.)

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 3
        assert np.all(res[[54, 55, 56]] == 1.)

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 4
        assert res[0] == 1.
        assert np.all(res[[54, 55, 56]] == 1.)

    def test_tau_from_topo_vect_all(self):
        self._aux_test_tau_from_topo_vect_all()

    def _aux_test_tau_from_topo_vect_all(self, proxy=None):
        # for the complete topology
        if proxy is None:
            proxy = self.proxy

        obs = self.env.reset()
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 0

        # the "first" topology: change the first element of the substation 0
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 1
        assert tau[0] == 1.

        # second topology: change the second element of substation 0.
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(0, (1, 2, 1))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 1
        assert tau[1] == 1.

        # the last topology identified: everything on bus 2 for substation 13 (last one)
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 1
        assert tau[389] == 1.

        # as there are 3 elements on the substation 0, there are 7 (=2^3 - 1) different possible topologies
        # for this one
        # And this will be the "first topology of substation 1"
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(1, (2, 1, 1, 1, 1, 1))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 1
        assert tau[7] == 1.

        # change the substation 12
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(12, (2, 2, 2, 2))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 1
        assert tau[382] == 1.

        # if i change 2 substation, i have 2 "1" on the tau vector
        obs = self.env.reset()
        act = self.env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = self.env.step(act)
        act = self.env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = self.env.step(act)
        tau = proxy.topo_vect_handler(obs)
        assert np.sum(tau) == 2
        assert tau[389] == 1.
        assert tau[0] == 1.

    def test_tau_from_list_topo(self):
        self. _aux_test_tau_from_list_topo()

    def _aux_test_tau_from_list_topo(self, proxy=None):
        if proxy is None:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="given_list",
                                 kwargs_tau=[(0, (2, 1, 1)), (0, (1, 2, 1)), (1, (2, 1, 1, 1, 1, 1)),
                                             (12, (2, 1, 1, 2)), (13, (2, 1, 2)), (13, (1, 2, 2)), 
                                             (1, (2, 1, 2, 1, 2, 1))
                                             ]
                                 )
            proxy.init([self.obs])

        env = self.env
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[0] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (1, 2, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[1] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 1, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[2] == 1.

        # I test that -1 are still considered as 1
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, -1, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        assert not done
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[2] == 1.

        # I test that -1 are still considered as 1, even if everything is on bus 1
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (1, -1, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        assert not done
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        # test that everything on bus 2 and everything on bus 1 leads to the same result
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(12, (2, 2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        # check cumulative "actions"
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 1, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 2
        assert res[0] == 1.
        assert res[4] == 1.

        # test the symmetry (again...) why not
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[2] == 1.
        
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(12, (1, 2, 2, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[3] == 1.

        # test that if a line is disconnected, we are still able to match the topologies
        # see https://github.com/BDonnot/leap_net/pull/12/files
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 1, -1, 1, 2, 1))]}})  # (2, 1, 2, 1, 2, 1)
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[6] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, -1, 2, 1, 2, 1))]}})  # (2, 1, 2, 1, 2, 1)
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[6] == 1.
        
    def test_tau_from_online_topo(self):
        self._aux_test_tau_from_online_topo()

    def _aux_test_tau_from_online_topo(self, proxy=None):
        if proxy is None:
            proxy = ProxyLeapNet(attr_tau=("line_status", "topo_vect",),
                                 topo_vect_to_tau="online_list",
                                 kwargs_tau=7
                                 )
            proxy.init([self.obs])

        env = self.env

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[0] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (1, 2, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[1] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 1, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[2] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(12, (2, 2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 2, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(0, (2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        act = env.action_space({"set_bus": {"substations_id": [(13, (2, 1, 2))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 2
        assert res[0] == 1.
        assert res[3] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[4] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 2, 1, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[5] == 1.

        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 1
        assert res[6] == 1.

        # "overflow" in the number of topologies => like ref topology
        obs = env.reset()
        act = env.action_space({"set_bus": {"substations_id": [(1, (2, 1, 1, 2, 1, 1))]}})
        obs, reward, done, info = env.step(act)
        with warnings.catch_warnings():
            # check that it raises the right warnings
            warnings.filterwarnings("error")
            with self.assertRaises(UserWarning):
                res = proxy.topo_vect_handler(obs)
        # check despite the warning, the things are processed correctly
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            res = proxy.topo_vect_handler(obs)
        assert np.sum(res) == 0



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.
import numpy as np
import warnings


def mape(y_true, y_pred, multioutput="uniform", threshold=1.):
    """
    Computes the "MAPE" norm (mean absolute percentage error).

    if multioutput is "uniform" then a single number is returned and it consists in taking the average of the previous
    vector.

    Notes
    ------
    This function completely ignores the values where `y_true` are such that `|y_true| < threshold`. It only considers
    values of y_true above the `threshold`.

    Parameters
    ----------
    y_true: ``numpy.ndarray``
        The true values. Each rows is an example, each column is a variable.

    y_pred: ``numpy.ndarray``
        The predicted values. Its shape should match the one from `y_true`

    multioutput: ``str``
        Whether or not to aggregate the returned values

    threshold: ``float``
        The the section "Notes" for more information. This should be a floating point number >= 0.

    Returns
    -------
    mape_: ``float`` or ``numpy.ndarray``
        If `multioutput` is "uniform" it will return a floating point number, otherwise it will return a vector
        with as many component as the number of columns in y_true an y_pred.

    """
    if y_true.shape != y_pred.shape:
        raise RuntimeError("mape can only be computed if y_true and y_pred have the same shape")

    try:
        threshold = float(threshold)
    except Exception as exc_:
        raise exc_
    if threshold < 0.:
        raise RuntimeError("The threshold should be a positive floating point value.")

    index_ok = (np.abs(y_true) > threshold)
    rel_error_ = np.full(y_true.shape, fill_value=np.NaN, dtype=float)
    rel_error_[index_ok] = (y_pred[index_ok] - y_true[index_ok])/y_true[index_ok]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        mape_ = np.nanmean(np.abs(rel_error_), axis=0)
    if multioutput == "uniform":
        mape_ = np.mean(mape_)
    return mape_



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

__all__ = ["RandomNN1", "RandomN2", "RandomN1",
           "RandomSub1", "RandomSub2", "RandomRefSub1",
           "RandomAct1", "RandomAct2", "RandomRefAct1"]

from leap_net.agents.randomNN1 import RandomNN1
from leap_net.agents.randomN1 import RandomN1
from leap_net.agents.randomN2 import RandomN2
from leap_net.agents.randomSub1 import RandomSub1
from leap_net.agents.randomSub2 import RandomSub2
from leap_net.agents.randomRefSub1 import RandomRefSub1
from leap_net.agents.randomAct1 import RandomAct1
from leap_net.agents.randomAct2 import RandomAct2
from leap_net.agents.randomRefAct1 import RandomRefAct1



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import multiply as tfk_multiply


class LtauNoAdd(Layer):
    """
    This layer implements the Ltau layer.

    This kind of leap net layer computes, from their input `x`: `d.(e.x * tau)` where `.` denotes the
    matrix multiplication and `*` the elementwise multiplication.

    Compare to a full Ltau block, this one does not add back the input.
    """

    def __init__(self, initializer='glorot_uniform', use_bias=True, trainable=True, name=None, **kwargs):
        super(LtauNoAdd, self).__init__(trainable=trainable, name=name, **kwargs)
        self.initializer = initializer
        self.use_bias = use_bias
        self.e = None
        self.d = None

    def build(self, input_shape):
        is_x, is_tau = input_shape
        nm_e = None
        nm_d = None
        if self.name is not None:
            nm_e = '{}_e'.format(self.name)
            nm_d = '{}_d'.format(self.name)
        self.e = Dense(is_tau[-1],
                       kernel_initializer=self.initializer,
                       use_bias=self.use_bias,
                       trainable=self.trainable,
                       name=nm_e)
        self.d = Dense(is_x[-1],
                       kernel_initializer=self.initializer,
                       use_bias=False,
                       trainable=self.trainable,
                       name=nm_d)

    def get_config(self):
        config = super().get_config().copy()
        config.update({
            'initializer': self.initializer,
            'use_bias': self.use_bias
        })
        return config

    def call(self, inputs, **kwargs):
        x, tau = inputs
        tmp = self.e(x)
        tmp = tfk_multiply([tau, tmp])  # element wise multiplication
        res = self.d(tmp)  # no addition of x
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

from leap_net.proxy.utils import DEFAULT_METRICS

from leap_net.ResNetLayer import ResNetLayer
from leap_net.proxy.evaluate_proxy_case_14 import main as main_14


def main(
        # reload the model
        env_name="l2rpn_neurips_2020_track2_small",
        save_path="model_saved_118",
        use_lightsim_if_available=True,
        val_regex=".*Scenario_february_0[0-9].*",
        model_name="leapnet_case_118",
        layer=ResNetLayer,  # for now this is not serialized in the model
        # parameters for the evaluation
        do_dc=True,
        do_N1 = True,
        do_N2 = True,
        li_batch_size=tuple(),  # if you want to study the impact of the batch size
        total_evaluation_step=int(1024) * int(128),
        pred_batch_size=int(1024) * int(32),
        save_path_final_results="model_results_118",  # where the information about the prediction will be stored
        metrics=DEFAULT_METRICS,  # which metrics are used to evaluate the performance of the model
        verbose=1,  # do I print the results of the model
        # enforce reproducibility
        chron_id_val=0,
        env_seed=0,
        agent_seed=42,
        ):
    main_14(env_name=env_name,
            save_path=save_path,
            use_lightsim_if_available=use_lightsim_if_available,
            val_regex=val_regex,
            model_name=model_name,
            layer=layer,
            do_dc=do_dc,
            do_N1=do_N1,
            do_N2=do_N2,
            li_batch_size=li_batch_size,
            total_evaluation_step=total_evaluation_step,
            pred_batch_size=pred_batch_size,
            save_path_final_results=save_path_final_results,
            metrics=metrics,
            verbose=verbose,
            chron_id_val=chron_id_val,
            env_seed=env_seed,
            agent_seed=agent_seed
            )


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

from grid2op.dtypes import dt_float
from leap_net.agents.randomAct1 import RandomAct1


class RandomRefAct1(RandomAct1):
    """
    This "agent" will randomly change the topology of one substation. The way the "randomly" works can be:

    - either it's random uniform on all the topologies (not recommended except for some tests)
    - or it will first draw a substation at random, and then perform a random action (uniformly) among this
      substation. We recommend this later version for normal grid (ieee118, real grid etc.)

    So basically we do not recommend to change the default "random_king='substation'"


    Notes
    -----
    This agent will modify all the substations at all steps. Make sure the `env.parameters.MAX_SUB_CHANGED` is
    big enough !

    Also the `env.parameters.NB_TIMESTEP_COOLDOWN_SUB` need to be large enough ! Otherwise a substation cannot be
    acted upon at every step.

    Same for `env.parameters.NB_TIMESTEP_COOLDOWN_LINE`

    """

    def __init__(self, action_space, p=0.5, list_act=()):
        super(RandomRefAct1, self).__init__(action_space, list_act=list_act)
        self.p = dt_float(p)
        self._1_p = 1. - self.p

    def act(self, obs, reward, done=False):
        ur = self.space_prng.uniform()
        if ur < self._1_p:
            res = self.all_to_one
        else:
            res = super().act(obs, reward, done)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import unittest
import grid2op
import warnings

import numpy as np

from leap_net.agents import RandomAct1


class TestRandomAct1(unittest.TestCase):
    def setUp(self) -> None:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            self.env = grid2op.make("l2rpn_case14_sandbox", test=True)

    def test_fail_if_nothing(self):
        with self.assertRaises(RuntimeError):
            agent = RandomAct1(self.env.action_space, list_act=[])

    def test_fail_if_only_do_nothing(self):
        with self.assertRaises(RuntimeError):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                agent = RandomAct1(self.env.action_space, list_act=[self.env.action_space()])

    def test_fail_if_not_glop_action(self):
        with self.assertRaises(RuntimeError):
            agent = RandomAct1(self.env.action_space, list_act=[None])

    def test_fail_if_only_one_sub(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}})
                    ]
        with self.assertRaises(RuntimeError):
            agent = RandomAct1(self.env.action_space, list_act=list_act)

    def test_doesnt_fail_if_not_only_one_sub(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]})
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)

    def test_can_sample_simple(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]})
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)
        for i in range(10):
            agent.act(None, None)

    def test_can_combine_topo_with_powerline(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]})
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)
        res = agent._combine_actions(list_act[0], list_act[2])
        assert np.sum(res.line_set_status == -1) == 1, "line 0 should be disconnected"
        assert res.line_set_status[0] == -1, "line 0 should be disconnected"
        *_, info = self.env.step(res)
        assert not info["is_illegal"], "this action should be legal"
        assert info["is_ambiguous"], "this action should not be ambiguous"

    def test_can_sample_more_complex_1(self):
        """i prevent to sample the topo action, it should sample the line one"""
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]})
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)
        act_id, sub_id_act, act = agent.sample_act(previous_act_id=2, sub_id_act=1)
        # check that it's the last action that is sampled
        assert act_id == 2
        assert sub_id_act is None
        assert np.sum(act.line_set_status == -1) == 1, "line 0 should be disconnected"
        assert act.line_set_status[0] == -1, "line 0 should be disconnected"

    def test_can_sample_more_complex_2(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]}),
                    self.env.action_space({"set_line_status": [(1, -1)]}),
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)
        act2 = 0
        act3 = 0
        for _ in range(20):
            act_id, sub_id_act, act = agent.sample_act(previous_act_id=1, sub_id_act=1)
            # check that it's the last action that is sampled
            assert act_id == 2 or act_id == 3
            if act_id == 2:
                act2 += 1
            else:
                act3 += 1
        # the test above is mainly to "check" that it samples randomly among the two possible action, and that
        # it does not always sample the same one
        assert act2 == 9
        assert act3 == 11

    def test_can_sample_more_complex_3(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 2, 1, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 1, 2, 1))]}})
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)
        act0 = 0
        act1 = 0
        act2 = 0
        act3 = 0
        # I sample uniformely on substation "before"
        for _ in range(20):
            act_id, sub_id_act, act = agent.sample_act(previous_act_id=2, sub_id_act=2)
            # check that it's the last action that is sampled
            assert act_id == 0 or act_id == 1
            if act_id == 0:
                act0 += 1
            else:
                act1 += 1
        # the test above is mainly to "check" that it samples randomly among the two possible action, and that
        # it does not always sample the same one
        assert act0 == 9
        assert act1 == 11

        # I sample uniformely on substation "after"
        for _ in range(20):
            act_id, sub_id_act, act = agent.sample_act(previous_act_id=2, sub_id_act=1)
            # check that it's the last action that is sampled
            assert act_id == 2 or act_id == 3
            if act_id == 2:
                act2 += 1
            else:
                act3 += 1
        # the test above is mainly to "check" that it samples randomly among the two possible action, and that
        # it does not always sample the same one
        assert act2 == 8
        assert act3 == 12

    def test_can_sample_more_complex_4(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 2, 1, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 1, 2))]}}),
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(0)

        act0 = 0
        act1 = 0
        act4 = 0
        act5 = 0
        # I sample uniformely on substation "after" or "before"
        for _ in range(1000):
            act_id, sub_id_act, act = agent.sample_act(previous_act_id=2, sub_id_act=2)
            # check that it's the last action that is sampled
            assert act_id == 0 or act_id == 1 or act_id == 4 or act_id == 5
            if act_id == 0:
                act0 += 1
            elif act_id == 1:
                act1 += 1
            elif act_id == 4:
                act4 += 1
            elif act_id == 5:
                act5 += 1
        # the test above is mainly to "check" that it samples randomly among the two possible action, and that
        # it does not always sample the same one
        assert act0 == 255
        assert act1 == 253
        assert act4 == 241
        assert act5 == 251

    def test_can_sample_more_complex_5(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 2, 1, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]}),
                    ]
        agent = RandomAct1(self.env.action_space, list_act=list_act)
        agent.seed(1)

        act0 = 0
        act1 = 0
        act4 = 0
        act5 = 0
        act6 = 0
        # I sample uniformely on substation "after" or "before"
        for _ in range(1000):
            act_id, sub_id_act, act = agent.sample_act(previous_act_id=2, sub_id_act=2)
            # check that it's the last action that is sampled
            assert act_id == 0 or act_id == 1 or act_id == 4 or act_id == 5 or act_id == 6 or act_id == 7
            if act_id == 0:
                act0 += 1
            elif act_id == 1:
                act1 += 1
            elif act_id == 4:
                act4 += 1
            elif act_id == 5:
                act5 += 1
            elif act_id == 6:
                act6 += 1
        # the test above is mainly to "check" that it samples randomly among the two possible action, and that
        # it does not always sample the same one
        assert act0 == 199
        assert act1 == 197
        assert act4 == 192
        assert act5 == 221
        assert act6 == 191



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import unittest
import grid2op
import warnings
import time
import numpy as np

from leap_net.agents import RandomAct2


class TestRandomAct2(unittest.TestCase):
    def setUp(self) -> None:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            self.env = grid2op.make("l2rpn_case14_sandbox", test=True)

    def test_can_sample_simple(self):
        list_act = [self.env.action_space({"set_bus": {"substations_id": [(1, (1, 2, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(1, (2, 2, 1, 1, 1, 2))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 2, 1, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(2, (2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 2, 1))]}}),
                    self.env.action_space({"set_bus": {"substations_id": [(3, (2, 1, 2, 1, 1, 2))]}}),
                    self.env.action_space({"set_line_status": [(0, -1)]}),
                    self.env.action_space({"set_line_status": [(1, -1)]}),
                    self.env.action_space({"set_line_status": [(2, -1)]}),
                    ]
        agent = RandomAct2(self.env.action_space, list_act=list_act)
        agent.seed(0)
        total_time = 0.
        nb_sim = 1000
        for i in range(nb_sim):
            act = agent.act(i, None)  # i pass the step in the act for debug, but it should be an observation otherwise
            beg_ = time.time()
            lines_impacted, _ = act.get_topological_impact()
            total_time += time.time() - beg_
            topo_vect = act.sub_set_bus
            nb_sub_impacted = 0
            prev = 0
            for sub_id, size_sub in enumerate(self.env.sub_info):
                this_sub_topo = topo_vect[prev: (prev + size_sub)]
                if np.sum(this_sub_topo == 2) > 0:
                    nb_sub_impacted += 1
                prev += size_sub
            assert np.sum(lines_impacted) + nb_sub_impacted == 2, f"action should act on two stuffs [step {i}]"
            # "2 stuffs" here is defined as:
            # - either two powerlines
            # - or two substations
            # - or a powerline and a substation

        # TODO if I have time, check the distribution of the simulated actions and make sure it's "uniform"



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.dtypes import dt_int, dt_float


class RandomN2(BaseAgent):
    """
    This "agent" will randomly disconnect exactly 2 different powerline from the grid.

    **NB** Every powerline that is not chosen at random to be disconnected will be reconnected by force.

    This agent will modify the status of all powerlines at every steps!

    Notes
    -----
    This agent will modify all the powerline at all steps. Make sure the `env.parameters.MAX_LINE_STATUS_CHANGED` is
    big enough !

    Also the `env.parameters.NB_TIMESTEP_COOLDOWN_LINE` need to be small enough ! Otherwise a powerline cannot be
    acted upon at every step.

    """

    def __init__(self, action_space):
        super(RandomN2, self).__init__(action_space)
        if "set_line_status" not in action_space.subtype.authorized_keys:
            raise NotImplementedError("Impossible to have a RandomN1 agent if you cannot set the status or powerline")

        # represent the action "exactly one powerline is disconnected
        self.powerline_actions = np.ones(action_space.n_line, dtype=dt_int)
        self.tmp_arr = np.ones(action_space.n_line, dtype=dt_int)

    def act(self, obs, reward, done=False):
        id_1 = self.space_prng.choice(self.action_space.n_line)
        id_2 = self.space_prng.choice(self.action_space.n_line - 1)
        if id_2 >= id_1:
            # this procedure is to be sure not to "disconnect twice" the same powerline
            id_2 += 1
        self.tmp_arr[:] = self.powerline_actions
        self.tmp_arr[id_1] = -1
        self.tmp_arr[id_2] = -1
        li_bus = [(i, el) for i, el in enumerate(self.tmp_arr)]
        act = self.action_space({"set_line_status": self.tmp_arr,
                                 "set_bus": {"lines_or_id": li_bus,
                                             "lines_ex_id": li_bus},
                                 }
                                )
        return act



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

from grid2op.dtypes import dt_float
from leap_net.agents.randomAct1 import RandomAct1


class RandomAct2(RandomAct1):
    """
    This combines two actions from `RandomAct1` each steps.

    The "rules" to combine the actions ensured that each action is composed of two sub actions and acts on two
    "stuffs" (either two powerlines, or two substation or 1 powerline and 1 substation).

    The rules are then the following:

    - two actions acting on the same substations cannot be composed together
    - the same action cannot be composed with itself.

    Notes
    -----
    This agent will modify all the substations at all steps. Make sure the `env.parameters.MAX_SUB_CHANGED` is
    big enough !

    Also the `env.parameters.NB_TIMESTEP_COOLDOWN_SUB` need to be large enough ! Otherwise a substation cannot be
    acted upon at every step.

    Same for `env.parameters.NB_TIMESTEP_COOLDOWN_LINE`
    """

    def __init__(self, action_space, list_act=()):
        super(RandomAct2, self).__init__(action_space, list_act=list_act)

    def act(self, obs, reward, done=False):
        act_id1, sub_id_act1, this_random1 = self.sample_act()
        if sub_id_act1 is not None:
            sub_id_act1 = sub_id_act1[0]
        act_id2, sub_id_act2, this_random2 = self.sample_act(act_id1, sub_id_act1)
        res = self._combine_actions(self.all_to_one, this_random1)
        res = self._combine_actions(res, this_random2)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import warnings
import numpy as np

from sklearn.metrics import mean_squared_error, mean_absolute_error  # mean_absolute_percentage_error
from leap_net.metrics import nrmse
from leap_net.metrics import pearson_r
from leap_net.metrics import mape
from leap_net.metrics import mape_quantile

import grid2op
from grid2op.Chronics import MultifolderWithCache
from grid2op.Parameters import Parameters


DEFAULT_METRICS = {"MSE_avg": mean_squared_error,
                   "MAE_avg": mean_absolute_error,
                   "NRMSE_avg": nrmse,
                   "pearson_r_avg": pearson_r,
                   "mape_avg": mape,
                   "rmse_avg": lambda y_true, y_pred: np.sqrt(mean_squared_error(y_true=y_true, y_pred=y_pred)),
                   "mape_90_avg": mape_quantile,
                   "MSE": lambda y_true, y_pred: mean_squared_error(y_true, y_pred, multioutput="raw_values"),
                   "MAE": lambda y_true, y_pred: mean_absolute_error(y_true, y_pred, multioutput="raw_values"),
                   "NRMSE": lambda y_true, y_pred: nrmse(y_true, y_pred, multioutput="raw_values"),
                   "pearson_r": lambda y_true, y_pred: pearson_r(y_true, y_pred, multioutput="raw_values"),
                   "mape": lambda y_true, y_pred: mape(y_true, y_pred, multioutput="raw_values"),
                   "rmse": lambda y_true, y_pred: np.sqrt(
                       mean_squared_error(y_true=y_true, y_pred=y_pred, multioutput="raw_values")),
                   "mape_90": lambda y_true, y_pred: mape_quantile(y_true=y_true,
                                                                   y_pred=y_pred,
                                                                   multioutput="raw_values"),
                   }


def limit_gpu_usage():
    import tensorflow as tf  # lazy import because the module "utils" can be imported from lots of different module
    # no need to import tensorflow each time
    physical_devices = tf.config.list_physical_devices('GPU')
    # Currently, memory growth needs to be the same across GPUs
    for device in physical_devices:
        tf.config.experimental.set_memory_growth(device, True)


def get_parameters():
    # generate the environments parameters
    param = Parameters()
    param.NO_OVERFLOW_DISCONNECTION = True
    param.NB_TIMESTEP_COOLDOWN_LINE = 0
    param.NB_TIMESTEP_COOLDOWN_SUB = 0
    param.MAX_SUB_CHANGED = 99999
    param.MAX_LINE_STATUS_CHANGED = 99999
    param.NB_TIMESTEP_COOLDOWN_SUB = 0
    return param


def create_env(env_name, use_lightsim_if_available=True):
    """create the grid2op environment with the right parameters and chronics class"""
    backend_cls = None
    if use_lightsim_if_available:
        try:
            from lightsim2grid.LightSimBackend import LightSimBackend
            backend_cls = LightSimBackend
        except ImportError as exc_:
            warnings.warn(f"You ask to use lightsim backend if it's available. But it's not available on "
                          f"your system (error \"{exc_}\").")

    if backend_cls is None:
        from grid2op.Backend import PandaPowerBackend
        backend_cls = PandaPowerBackend

    param = get_parameters()

    env = grid2op.make(env_name,
                       param=param,
                       backend=backend_cls(),
                       chronics_class=MultifolderWithCache
                       )
    return env


def reproducible_exp(env, agent, env_seed=None, chron_id_start=None, agent_seed=None):
    """
    ensure the reproducibility for the data, but NOT for tensorflow

    the environment need to be reset after a call to this method
    """
    if env_seed is not None:
        env.seed(env_seed)

    if chron_id_start is not None:
        # set the environment to start at the right chronics
        env.chronics_handler.tell_id(chron_id_start - 1)

    if agent_seed is not None:
        agent.seed(agent_seed)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.
import numpy as np


def nrmse(y_true, y_pred, multioutput="uniform", threshold=1.):
    """
    Computes the "normalized RMSE" norm.

    This norms compute the mean squared error (average over the rows of the square of the difference between y_true
    and y_pred). This gives a vector with as many elements as the number of columns of y_true and y_pred.

    Then it computes the RMSE by taking the square root of each numbers.

    Then, this rmse vector is divided by the the 'max - min' over each column, which is the "normalization".

    if multioutput is "uniform" then a single number is returned and it consists in taking the average of the previous
    vector.

    Notes
    ------
    If the difference 'max - min' for a column is bellow the threshold (by default 1.) it is replaced by this threshold.
    This has been implemented to avoid numerical instability when some columns are almost equal everywhere (this can
    happen in the power system settings: for example some voltages are always constant). You can put "0." to
    deactivate this feature.

    The shapes of y_true and y_pred should match.

    Parameters
    ----------
    y_true: ``numpy.ndarray``
        The true values. Each rows is an example, each column is a variable.

    y_pred: ``numpy.ndarray``
        The predicted values. Its shape should match the one from `y_true`

    multioutput: ``str``
        Whether or not to aggregate the returned values

    threshold: ``float``
        The the section "Notes" for more information. This should be a floating point number >= 0.

    Returns
    -------
    nrmse_: ``float`` or ``numpy.ndarray``
        If `multioutput` is "uniform" it will return a floating point number, otherwise it will return a vector
        with as many component as the number of columns in y_true an y_pred.

    """
    if y_true.shape != y_pred.shape:
        raise RuntimeError("nrmse can only be computed if y_true and y_pred have the same shape")

    try:
        threshold = float(threshold)
    except Exception as exc_:
        raise exc_
    if threshold < 0.:
        raise RuntimeError("The threshold should be a positive floating point value.")

    se_ = (y_true - y_pred)**2
    mse = np.mean(se_, axis=0)
    rmse = np.sqrt(mse)
    norm_ = (np.max(y_true, axis=0) - np.min(y_true, axis=0))
    norm_[norm_ <= threshold] = threshold
    nrmse_ = rmse / norm_
    if multioutput == "uniform":
        nrmse_ = np.mean(nrmse_)
    return nrmse_



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import os
import time
import copy
import numpy as np
from abc import ABC, abstractmethod
from collections.abc import Iterable


class BaseProxy(ABC):
    """
    Base class you have to implement if you want to use easily a proxy.

    So example of such a class are given in `ProxyBackend` and `ProxyLeapNet`

    Attributes
    ----------

    name: ``str``
        The name of the proxy (used for example when saving or restoring data)

    max_row_training_set: ``int``
        An integer > 0. It represents the maximum of rows there will be in the "training set" of our proxy.

    eval_batch_size: ``int``
        An integer > 0. It represents the batch size of the proxy when it is used to make predictions. Sometimes
        for computation time reasons, it can be useful to batch multiple data and predict more than one state at
        the same time.

    dtype: ``type``
        Type of data used to store / retrieve it. By default it's `np.float32`

    last_id: ``int``
        The id in the database :attr:`_my_x` and :attr:`_my_y` corresponding to the last element entered.

    _global_iter: ``int``
        Total number of data received

    __first_eval: ``bool``
        Has the model been used to make prediction or not.

    _model: ``anything``
        The implementation of the proxy, it can be a grid2op backend, or a tensorflow model for example

    _time_predict: ``float``
        Time spent (measured in seconds) to make predictions with the proxy, only counting the evaluation time and
        not the time spent in acquiring the data. It is measure in seconds (s).

    _time_train: ``float``
        Time spent (measured in seconds) to train the proxy, only counting the training time and not the time
        spent in acquiring the data.

    attr_x: ``list`` of ``str``
        Name of the attribute sof the observation that are used as input to the proxy

    attr_y: ``list`` of ``str``
        Name of the attributes of the observation that are predicted by the proxy

    _my_x: ``list`` of ``numpy.ndarray``
        This represents the training dataset that is filled "on the fly" by running grid2op environment. This part
        of the dataset is the input of the proxy. We don't recommend to modify this

    _my_y: ``list`` of ``numpy.ndarray``
        This represents the training dataset that is filled "on the fly" by running grid2op environment. This part
        of the dataset is the real output of what the proxy should predict. We don't recommend to modify this

    _sz_x: ``list`` of ``int``
        For each input of the proxy, it gives its size. We don't recommend to modify this

    _sz_y: ``list`` of ``int``
        For each output of the proxy, it gives its size. We don't recommend to modify this

    _metadata_loaded: ``bool``
        Whether the metadata of the model have been loaded from hard drive, or created at the initialization of the
        model.

    _last_id_eval: ``int``
        Internal, do not use

    __db_full: ``bool``
        Internal, do not use

    """
    DEBUG = False

    def __init__(self,
                 name,
                 max_row_training_set=int(1e5),
                 eval_batch_size=1,
                 attr_x=("prod_p", "prod_v", "load_p", "load_q", "topo_vect"),  # input that will be given to the proxy
                 attr_y=("a_or", "a_ex", "p_or", "p_ex", "q_or", "q_ex", "prod_q", "load_v", "v_or", "v_ex"),  # output that we want the proxy to predict
                 ):
        # name
        self.name = name

        # data type
        self.dtype = np.float32

        # to fill the training / test dataset
        self.max_row_training_set = max_row_training_set
        self.eval_batch_size = eval_batch_size
        if self.max_row_training_set < self.eval_batch_size:
            raise RuntimeError(f"You cannot use a batch size of {self.eval_batch_size} with a dataset counting at"
                               f" most {self.max_row_training_set} rows. "
                               "Please either increase \"max_row_training_set\" or decrease \"batch_size\""
                               "(hint: batch_size>=max_row_training_set).")

        # database handling
        self.last_id = 0  # last index in the database
        self._global_iter = 0  # total number of data received
        self.__db_full = None  # is the "training database" full
        self.__first_eval = True

        # the model
        self._model = None

        # timers
        self._time_predict = 0
        self._time_train = 0

        # for the prediction
        self._last_id_eval = 0

        # database part
        self.attr_x = attr_x
        self.attr_y = attr_y
        self._my_x = None
        self._my_y = None
        self._sz_x = None
        self._sz_y = None
        self._metadata_loaded = False

    #######################################################################
    ## All functions bellow should be implemented in your specific proxy ##
    #######################################################################
    @abstractmethod
    def build_model(self):
        """
        build the model that is used inside of the proxy

        Be careful this function can be called multiple times

        """
        pass

    @abstractmethod
    def _make_predictions(self, data, training=False):
        """
        Make a prediction with the proxy on a new grid state.

        It's analogous to the `train_model` but instead of training it gives the prediction of the neural network.

        It's called when the proxy is being evaluated with:

        .. code-block:: python

            data = self._extract_data([last_index])
            tmp = self.make_predictions(data)
            res = self._post_process(tmp)

        It's not part of the public API that is used outside of your proxy (private method), and it is only used
        with the given code above.

        Parameters
        ----------
        data: ``list`` of ``float``
            The list of arrays corresponding to each attribute in `attr_x`: the input of the proxy.

        training: ``bool``
            Whether the model is being trained or not

        Returns
        -------
        res: ``list`` of ``float``
            The list of the predictions made by the proxy: each element of the list
            corresponding to the prediction of the variable in `attr_y` - so each element is a vector- made with
            input in `data`
        """
        pass

    #######################################################################
    ## All functions bellow can be implemented in your specific but if   ##
    ## this is the case we recommend to call the method of this base class ##
    ########################################################################
    def init(self, obss):
        """
        initialize the meta data needed for the model to run (obss is a list of observations)

        This function may be overridden but in that case we recommend to call the method of the super class.

        If it's overridden, we also recommend to call `_init_database_shapes`

        Parameters
        ----------
        obss: ``list`` of ``grid2op.Observation``
            List of observations used to inialize this model, for example on which the model will compute the mean
            and standard deviation to scale the data.

        """

        self.__db_full = False
        obs = obss[0]
        # save the input x
        self._my_x = []
        self._sz_x = []
        self._my_y = []
        self._sz_y = []

        # init the dimension of everything
        for attr_nm in self.attr_x:
            arr_ = self._extract_obs(obs, attr_nm)
            sz = arr_.size
            self._sz_x.append(sz)
        for attr_nm in self.attr_y:
            arr_ = self._extract_obs(obs, attr_nm)
            sz = arr_.size
            self._sz_y.append(sz)
        self._init_database_shapes()

    def _init_database_shapes(self):
        """
        init the database to have the proper size.

        This function is automatically called by the "init" method or by the "load_metadata" method.

        This function may be overridden but in that case we recommend to call the method of the super class
        """
        # init the database
        self._my_x = []
        for sz in self._sz_x:
            self._my_x.append(np.zeros((self.max_row_training_set, sz), dtype=self.dtype))
        self._my_y = []
        for sz in self._sz_y:
            self._my_y.append(np.zeros((self.max_row_training_set, sz), dtype=self.dtype))

    def store_obs(self, obs):
        """
        This method update all the intermediate for you.

        You need to make a derivate function and call "super().store_obs()"
        to benefit at maximum from this Proxy interface.

        This function may be overridden but in that case we recommend to call the method of the super class

        Parameters
        ----------
        obs: ``grid2op.Observation``
            The observation to store in the database.

        """
        # update the counters
        # store only the first batch if DEBUG otherwise store all
        for attr_nm, inp in zip(self.attr_x, self._my_x):
            inp[self.last_id, :] = self._extract_obs(obs, attr_nm)
        for attr_nm, inp in zip(self.attr_y, self._my_y):
            inp[self.last_id, :] = self._extract_obs(obs, attr_nm)

        self._global_iter += 1
        self.last_id += 1
        if self.last_id >= self.max_row_training_set - 1:
            self.__db_full = True
        self.last_id %= self.max_row_training_set

    def load_metadata(self, dict_):
        """
        this function is used when loading the proxy to restore the meta data

        This function may be overridden but in that case we recommend to call the method of the super class

        Notes
        -----
        This function is expected to modify the instance on which it is called (*ie* `self`)

        Parameters
        ----------
        dict_: ``dict``
            The dictionary of parameter that is used to initialize this instance.

        """
        self.attr_x = tuple([str(el) for el in dict_["attr_x"]])
        self.attr_y = tuple([str(el) for el in dict_["attr_y"]])

        self._sz_x = [int(el) for el in dict_["_sz_x"]]
        self._sz_y = [int(el) for el in dict_["_sz_y"]]

        self._time_train = float(dict_["_time_train"])
        self._time_predict = float(dict_["_time_predict"])

        self._init_database_shapes()

        self._metadata_loaded = True

    def get_metadata(self):
        """
        should return a dictionary containing all the metadata of this class in a format that is compatible
        with json serialization.

        This function may be overridden but in that case we recommend to call the method of the super class

        Notes
        -----
        We assume that the metadata are "json serializable".

        Returns
        --------
        res: ``dict``
            A dictionary containing the necessary information to initialize the class as when it was saved.

        """
        res = dict()
        res["attr_x"] = [str(el) for el in self.attr_x]
        res["attr_y"] = [str(el) for el in self.attr_y]
        res["_sz_x"] = [int(el) for el in self._sz_x]
        res["_sz_y"] = [int(el) for el in self._sz_y]

        res["_time_train"] = float(self._time_train)
        res["_time_predict"] = float(self._time_predict)

        return res

    #######################################################################
    ##  All functions bellow can be implemented in your specific proxy   ##
    #######################################################################
    def load_data(self, path, ext=".h5"):
        """
        You need to override this function if the proxy, in order to be functional need to load some data
        that are not loaded by the `load_metadata` function.

        This function may be overridden but in that case we recommend to call the method of the super class

        load the weight of the neural network
        path is the full path (including file name and extension).

        This function can be overridden (for example if your proxy does not use tensorflow or is made of multiple
        submodule that need to be saved independently)

        This function is used when loading back your proxy.

        Notes
        -----
        This function is only called when the metadata (number of layer, size of each layer etc.)
        have been properly restored (so it supposes load_metadata has been successfully called)

        We suppose that there is a "." preceding the extension. So ext=".h5" is valid, but not ext="h5"

        We recommend that the file at the "path" is not directly read.
        Rather, it is better to first copied it into a temporary directory, and then
        read it. This is to avoid any data corruption when an instance of the model is reading and another
        is writing to the same file.

        """
        pass

    def save_data(self, path, ext=".h5"):
        """
        Save extra information that might be required by the model. For example this saves the weights of
        some neural network if needed.

        This function should be overridden

        Notes
        -----
        We suppose that there is a "." preceding the extension. So ext=".h5" is valid, but not ext="h5"

        Parameters
        ----------
        path: ``str``
            The path at which the model will be saved
        ext: ``str``
            The extension used to save the model (for example ".h5" should output a file named xxx.h5)

        """
        pass

    def _extract_data(self, indx_train):
        """
        extract from the training dataset, the data with indexes `indx_train`

        The model will be trained with a code equivalent to:

        .. code-block:: python

            data = self._extract_data(indx_train)
            batch_losses = self._train_model(data)

        This function is also used for the evaluation of the model in the following manner:

        .. code-block:: python

            data = self._extract_data(indx_val)
            res = self._make_predictions(data, training=False)

        This function can be overridden, but we don't necessarily recommend to do so

        Parameters
        ----------
        indx_train: ``numpy.ndarray``, ``int``
            The index of the data that needs to be retrieved from the database `_my_x` and `_my_y`

        Returns
        -------
        X:
            The value of the input data
        Y:
            The value of the desired output of the proxy

        """

        tmpx = [arr[indx_train, :] for arr in self._my_x]
        tmpy = [arr[indx_train, :] for arr in self._my_y]
        return tmpx, tmpy

    def get_output_sizes(self):
        """
        Should return the list of the dimension of the output of the proxy.

        This function can be overridden

        Returns
        -------
        res: ``list`` of ``int``
            The list representing the sizes of each output
        """
        return copy.deepcopy(self._sz_y)

    def get_attr_output_name(self, obs):
        """
        Get the name (that will be used when saving the model) of each ouput of the proxy.

        This function may be overridden

        Parameters
        ----------
        obs: ``grid2op.BaseObservation``
            The observation (unique) from which to retrieve the data

        Returns
        -------
        res:
            The list of the attribute name (note that the observation might not be used at all)
        """
        return copy.deepcopy(self.attr_y)

    def get_true_output(self, obs):
        """
        Returns, from the observation the true output that has been computed by the environment.

        This "true output" is computed based on the observation and corresponds to what the proxy is meant to
        approximate (but the reference)

        This function may be overridden

        Parameters
        ----------
        obs: ``grid2op.BaseObservation``
            The observation (unique) from which to retrieve the data

        Returns
        -------
        res: ``list of float``
            The real values of each output as extracted from the observation.

        """
        res = []
        for attr_nm in self.attr_y:
            res.append(self._extract_obs(obs, attr_nm))
        return res

    def save_tensorboard(self, tf_writer, training_iter, batch_losses):
        """
        Save the information of the last training iteration in tensorboard.

        This function may be overridden

        Parameters
        ----------
        tf_writer:
            Tensorflow writer on which to write the data

        training_iter: ``int``
            Current training iteration

        batch_losses: ``list of float``
            The losses for each variables that the proxy is supposed to predict

        """
        pass

    def _post_process(self, predicted_state):
        """
        This function is used to post process the data that are the output of the proxy.

        For example, this can be used by a proxy using a neural network, if some scaling have been done, to
        "unscale" the predictions

        This function can be overridden

        Parameters
        ----------
        predicted_state: ``list`` of ``float``
            For each variables, it contains the (raw) predictions of the proxy

        Returns
        -------
        res: ``list`` of ``float``
            For each variables, it should return the post processed values.

        """
        return predicted_state

    def _extract_obs(self, obs, attr_nm):
        """
        Extract a given attribute from an observation.

        This function can be overridden if you want to extract non attribute (for example results of a method) of
        an observation.

        Parameters
        ----------
        obs:
            Grid2op action on which we want to extract something

        attr_nm:
            Name of the attribute we want to extract

        Returns
        -------
        res:
            The array representing what need to be extracted from the observation
        """
        return getattr(obs, attr_nm)

    #######################################################
    ## We don't recommend to change anything bellow this ##
    #######################################################
    def train(self, tf_writer=None):
        """
        Train the proxy (if tf_writer is not None, it is expected that the proxy save the computation graph

        If your model need to be trained, we recommend you to inherit from "BaseNNProxy" instead

        Parameters
        ----------
        tf_writer

        Returns
        -------
        None if the proxy has not been trained at this iteration, or the losses

        """
        raise RuntimeError("This model cannot be trained!")

    def predict(self, force=False):
        """
        Make the prediction using the proxy. Predictions are made using the last "eval_batch_size" data present on the
        database.

        The way it's currently implemented is that this function returns ``None`` if not enough data have been
        gathered since the last call to this function.

        We do not recommend to override this function.

        Parameters
        ----------
        force: ``bool``
            Whether to force the predictions or not

        Notes
        -----
        It can only be called if the proxy has been "filled" with the observations before.

        If the "prediction batch size" is not 1, this function will sometimes return ``None``.

        Returns
        -------
        res:
            All the predictions made by the proxy. Note that these predictions should be directly comparable with
            the input data and (so if a scaling is applied, they should be unscaled).

            res is sometimes ``None``

        """
        if (self._global_iter % self.eval_batch_size != 0) and (not force):
            return None
        data = self._extract_data(np.arange(self._last_id_eval, self._global_iter) % self.max_row_training_set)

        if self.__first_eval:
            # evaluate at "blank" the first time so that tensorflow / keras can load the model
            res = self._make_predictions(data, training=False)
            self.__first_eval = False

        beg_ = time.time()
        res = self._make_predictions(data, training=False)
        self._time_predict += time.time() - beg_
        res = self._post_process(res)
        self._last_id_eval = self._global_iter
        return res

    def _save_dict(self, li, val):
        """
        save the metadata of this proxy into a valid representation of `self`. It's a utility function to convert
        into "json serializable object" the numpy data.

        It is a helper to convert data in float format from either a list or a single numpy floating point.

        We don't recommend to override this function.

        Parameters
        ----------
        li
        val

        Returns
        -------

        """
        if isinstance(val, Iterable):
            li.append([float(el) for el in val])
        else:
            li.append(float(val))

    def _add_attr(self, attr_nm, val):
        """
        add an attribute to myself based on the value (that can either be a list or a single element)
        used in "from_dict" for example

        We don't recommend to override this function.

        """
        if isinstance(val, Iterable):
            getattr(self, attr_nm).append(np.array(val).astype(self.dtype))
        else:
            getattr(self, attr_nm).append(self.dtype(val))

    def _get_adds_mults_from_name(self, obss, attr_nm):
        """
        extract the scalers (mean and std) used for the observation

        We don't recommend to overide this function, modify the function `_get_mean` and `_get_sd` instead

        obss is a list of observation obtained from running some environment with just the "actor" acting on
        the grid. The size of this list is set by `AgentWithProxy.nb_obs_init`

        Notes
        ------
        for each variables, the data are scaled with:

        data_scaled = (data_raw - add_tmp) / mult_tmp

        """
        obs = obss[0]
        add_tmp = np.mean([self._extract_obs(ob, attr_nm) for ob in obss], axis=0).astype(self.dtype)
        mult_tmp = np.std([self._extract_obs(ob, attr_nm) for ob in obss], axis=0).astype(self.dtype) + 1e-1

        if attr_nm in ["prod_p"]:
            # mult_tmp = np.array([max((pmax - pmin), 1.) for pmin, pmax in zip(obs.gen_pmin, obs.gen_pmax)],
            #                     dtype=self.dtype)
            # default values are good enough
            pass
        elif attr_nm in ["prod_q"]:
            # default values are good enough
            pass
        elif attr_nm in ["load_p", "load_q"]:
            # default values are good enough
            pass
        elif attr_nm in ["load_v", "prod_v"]:
            # default values are good enough
            # stds are almost 0 for loads, this leads to instability
            add_tmp = np.mean([self._extract_obs(ob, attr_nm) for ob in obss], axis=0).astype(self.dtype)
            mult_tmp = 1.0  # np.mean([self._extract_obs(ob, attr_nm) for ob in obss], axis=0).astype(self.dtype)
        elif attr_nm in ["v_or", "v_ex"]:
            # default values are good enough
            add_tmp = self.dtype(0.)  # because i multiply by the line status, so i don't want any bias
            mult_tmp = np.mean([self._extract_obs(ob, attr_nm) for ob in obss], axis=0).astype(self.dtype)
        elif attr_nm == "hour_of_day":
            add_tmp = self.dtype(12.)
            mult_tmp = self.dtype(12.)
        elif attr_nm == "minute_of_hour":
            add_tmp = self.dtype(30.)
            mult_tmp = self.dtype(30.)
        elif attr_nm == "day_of_week":
            add_tmp = self.dtype(4.)
            mult_tmp = self.dtype(4)
        elif attr_nm == "day":
            add_tmp = self.dtype(15.)
            mult_tmp = self.dtype(15.)
        elif attr_nm in ["target_dispatch", "actual_dispatch"]:
            add_tmp = self.dtype(0.)
            mult_tmp = np.array([max((pmax - pmin), 1.) for pmin, pmax in zip(obs.gen_pmin, obs.gen_pmax)],
                                dtype=self.dtype)
        elif attr_nm in ["p_or", "p_ex", "q_or", "q_ex"]:
            add_tmp = self.dtype(0.)  # because i multiply by the line status, so i don't want any bias
            mult_tmp = np.array([max(np.abs(val), 1.0) for val in getattr(obs, attr_nm)], dtype=self.dtype)
        elif attr_nm in ["a_or", "a_ex"]:
            add_tmp = self.dtype(0.)  # because i multiply by the line status, so i don't want any bias
            if hasattr(obs, "thermal_limit"):
                mult_tmp = 1.0 * obs.thermal_limit
            elif hasattr(obs, "rho"):
                mult_tmp = np.abs(obs.a_or / (obs.rho + 1e-2))  # which is equal to the thermal limit
            else:
                # defaults would be good enough
                pass
            mult_tmp[mult_tmp <= 1.] = 1.0
        elif attr_nm == "line_status":
            # encode back to 0: connected, 1: disconnected
            add_tmp = self.dtype(1.)
            mult_tmp = self.dtype(-1.0)

        return add_tmp, mult_tmp

    def get_total_predict_time(self):
        """
        get the total time spent to make the prediction with the proxy

        We don't recommend to overide this function
        """
        return self._time_predict

    def _is_db_full(self):
        return self.__db_full



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.dtypes import dt_int, dt_float


class RandomSub1(BaseAgent):
    """
    This "agent" will randomly change the topology of one substation. The way the "randomly" works can be:

    - either it's random uniform on all the topologies (not recommended except for some tests)
    - or it will first draw a substation at random, and then perform a random action (uniformly) among this
      substation. We recommend this later version for normal grid (ieee118, real grid etc.)

    So basically we do not recommend to change the default "random_king='substation'"

    Notes
    -----
    This agent will modify all the substations at all steps. Make sure the `env.parameters.MAX_SUB_CHANGED` is
    big enough !

    Also the `env.parameters.NB_TIMESTEP_COOLDOWN_SUB` need to be large enough ! Otherwise a substation cannot be
    acted upon at every step.

    """

    def __init__(self, action_space, random_kind="substation"):
        super(RandomSub1, self).__init__(action_space)
        if "set_bus" not in action_space.subtype.authorized_keys:
            raise NotImplementedError("Impossible to have a RandomSub1 agent if you cannot set the bus of the "
                                      "substations")

        # represent the action "exactly one powerline is disconnected
        self.all_to_one = action_space()
        self.all_to_one.set_bus = np.ones(action_space.dim_topo, dtype=dt_int)
        self.all_topo = []
        self.random_kind = random_kind
        self.sub_considered = []  # list of substation with more than 2 feasible topology
        self.sub_to_possible_id = np.zeros(action_space.n_sub, dtype=int)
        possible_id = 0
        for sub_id in range(action_space.n_sub):
            tmp = action_space.get_all_unitary_topologies_set(action_space, sub_id)
            if len(tmp):
                act_non_dn = tmp[1:]
                if self.random_kind == "substation":
                    self.sub_considered.append(sub_id)
                    self.all_topo.append((sub_id, act_non_dn))  # don't take into account the do nothing action
                    self.sub_to_possible_id[sub_id] = possible_id
                    possible_id += 1
                elif self.random_kind == "all":
                    self.all_topo += act_non_dn
                else:
                    raise RuntimeError(f"Unknown sampling type: \"{random_kind}\"")

    def sample_act(self, previous_sub_id=None):
        sub_id = None
        if self.random_kind == "substation":
            if previous_sub_id is None:
                id_ = self.space_prng.choice(len(self.all_topo))
            else:
                id_ = self.space_prng.choice(len(self.all_topo) - 1)
                if id_ >= self.sub_to_possible_id[previous_sub_id]:
                    # this procedure is to be sure not to "disconnect twice" the same powerline
                    id_ += 1
            sub_id, all_topo_this_sub = self.all_topo[id_]
            this_random = self.space_prng.choice(all_topo_this_sub)
        elif self.random_kind == "all":
            this_random = self.space_prng.choice(self.all_topo)
        else:
            raise RuntimeError(f"Unknown sampling type: \"{self.random_kind}\"")
        return sub_id, this_random

    def act(self, obs, reward, done=False):
        sub_id, this_random = self.sample_act()
        res = self.all_to_one + this_random
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.
import numpy as np
from scipy.stats import pearsonr


def pearson_r(y_true, y_pred, multioutput="uniform", threshold=1.0):
    """
    Compute the pearson correlation coefficient given two matrices y_true and y_pred.

    Each row is supposed to be an example (on which the pearson correlation coefficient) is computed. The computation
    happens independently for each column.

    Notes
    ------
    If the difference 'max - min' for a column is bellow the threshold (by default 1.) this routines consider that
    the value is constant and thus do not compute the pearson correlation coefficient and returns "nan" instead.

    When multioutput is not "uniform" the average is computed only on the "non nan" variables. Nan component are
    ignored.

    Parameters
    ----------
    y_true: ``numpy.ndarray``
        The true values. Each rows is an example, each column is a variable.

    y_pred: ``numpy.ndarray``
        The predicted values. Its shape should match the one from `y_true`

    multioutput: ``str``
        Whether or not to aggregate the returned values

    threshold: ``float``
        The the section "Notes" for more information. This should be a floating point number >= 0.

    Returns
    -------
    rs: ``float`` or ``numpy.ndarray``
        If `multioutput` is "uniform" it will return a floating point number corresponding to the average pearson
        correlation coefficient, otherwise it will return a vector with as many component as the number of columns
        in y_true an y_pred.
    """
    if y_true.shape != y_pred.shape:
        raise RuntimeError("pearson_r can only be computed if y_true and y_pred have the same shape")
    if len(y_true.shape) != 2:
        raise RuntimeError("pearson_r can only be used with matrices")

    rs = np.zeros(y_true.shape[1])
    for col_id in range(y_true.shape[1]):
        true_tmp = y_true[:, col_id]
        pred_tmp = y_pred[:, col_id]

        # don't count some value considered "constant"
        is_ko_true = np.abs(np.max(true_tmp) - np.min(true_tmp)) <= threshold
        is_ko_pred = np.abs(np.max(pred_tmp) - np.min(pred_tmp)) <= threshold
        if is_ko_true or is_ko_pred:
            tmp_r = np.NaN
        else:
            tmp_r, *_ = pearsonr(true_tmp, pred_tmp)
        rs[col_id] = tmp_r

    if multioutput == "uniform":
        if np.all(~np.isfinite(rs)):
            # if everything is Nan then the returned value is Nan
            rs = np.NaN
        else:
            # don't take into account the nan there
            rs = np.nanmean(rs)
    return rs



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import copy
import warnings
import numpy as np
from typing import Union

from grid2op.Agent import BaseAgent
from grid2op.Action import BaseAction
from grid2op.dtypes import dt_int, dt_float


class RandomAct1(BaseAgent):
    """
    This agent will send random action from a provided list (provided with argument
    `list_act`)

    The actions should meet some properties:

    - each action should be unique (not checked, but expect errors if this is not met)
    - taken as a whole, actions should act on different substation (meaning that either no action act on any substation
      or there are 2 different actions acting on two different substation)
    - each action should act on at most one substation
    - do nothing action will not be added (but it's not an error)
    - actions should use "set" and not "change" (ie `set_bus` and `set_line_status` but NOT `change_bus` nor
      `change_line_status`)
    - at least one action should do something (full do nothing action or an empty list are not allowed)
    - disconnection of element using "set_bus" are not allowed (not checked but expect some bugs if you do that)
    - should set all the elements of a substation if it acts on it (ie should explicitely set to 1 and not keep
      at 0 some elements)

    Notes
    -----
    This agent will modify all the substations at all steps. Make sure the `env.parameters.MAX_SUB_CHANGED` is
    big enough !

    Also the `env.parameters.NB_TIMESTEP_COOLDOWN_SUB` need to be large enough ! Otherwise a substation cannot be
    acted upon at every step.

    Same for `env.parameters.NB_TIMESTEP_COOLDOWN_LINE`
    """

    def __init__(self, action_space, list_act=()):
        super(RandomAct1, self).__init__(action_space)
        if "set_bus" not in action_space.subtype.authorized_keys:
            raise NotImplementedError("Impossible to have a RandomSub1 agent if you cannot set the bus of the "
                                      "substations")

        # represent the action "exactly one powerline is disconnected
        self.all_to_one = action_space()
        self.all_to_one.set_bus = np.ones(action_space.dim_topo, dtype=dt_int)
        self._all_actions = []
        self._subs_impacted_by_action = []
        self._n_sub = self.action_space.n_sub
        self._affect_lines = self.action_space.n_sub
        self._nb_act = np.zeros(self._n_sub + 1, dtype=int)  # [sub_0, sub_1, ..., sub_n, powerlines]

        # self._subs_impacted_to_act_id:
        #     sub_0: {id of actions acting on sub_0 (and possibly other sub)},
        #     sub_1: {id of actions acting on sub 1 (and possibly other sub)},
        #     ..., sub_n: {...},
        #     powerlines: {id of actions acting only on powerlines}
        self._subs_impacted_to_act_id = [[] for _ in range(self._n_sub + 1)]

        # add the actions in the list, and perform lots of checks! (to avoid misuse)
        subs_impact = np.zeros(action_space.n_sub, dtype=int)
        warning_different_act_same_sub_issued = False
        all_act_topo = True
        act_id = 0
        for act_ref_id, act in enumerate(list_act):
            if not isinstance(act, BaseAction):
                # act should be a grid2op actions...
                raise RuntimeError(f"The list_act should contain only valid grid2op actions. Found {type(act)}.")
            if np.any(act.line_change_status):
                # act should not "change" the status of powerline but "set" them
                raise RuntimeError("action of type \"line_change_status\" are not supported at the moment. Please use "
                                   "`line_change_status`.")
            if np.any(act.change_bus):
                # act should not "change" the bus of elements but "set" them
                raise RuntimeError("action of type \"change_bus\" are not supported at the moment please use "
                                   "`change_bus`.")

            # TODO do i check for "-1" in the set_bus ?
            # TODO do i check for "0" in the "set_bus" of the affected substation ?

            # all below is to make sure that, when combining the action, i actually result in the
            # combined actions and not just the second one (this would be the case if i combine two actions
            # on the same substation)
            lines_impacted, subs_impacted = act.get_topological_impact()
            if np.sum(lines_impacted) + np.sum(subs_impacted) == 0:
                # action is do nothing
                warnings.warn("One of the provided action is do nothing. We will NOT add it there.")
                continue

            if np.sum(subs_impacted) > 1:
                warnings.warn(f"For now RandomAct1 (and especially RandomAct2) only works when provided actions "
                              f"act each on at most one substation. Please check element {act_ref_id} of the "
                              f"provided action")
            subs_id = np.where(subs_impacted)[0]
            if np.any(subs_impact[subs_id] > 1):
                if not warning_different_act_same_sub_issued:
                    subs_this = subs_id[np.where(subs_impact[subs_id] > 1)[0]]
                    warnings.warn(f"More than one unitary action affect substation: {subs_this}. Note that when "
                                  f"combining "
                                  f"actions, the actions acting on the same substation will not be combined "
                                  f"together, as "
                                  f"it would NOT result in a combined action otherwise.")
                    warning_different_act_same_sub_issued = True

            # TODO check that action is not already added, maybe ?

            subs_impact[subs_id] += 1
            # explicitly put `None` if nothing is modified
            self._subs_impacted_by_action.append(subs_id if subs_id.size > 0 else None)
            if subs_id.size == 0:
                # i found an action not acting on a substation
                all_act_topo = False
                self._nb_act[self._affect_lines] += 1
                self._subs_impacted_to_act_id[self._affect_lines].append(act_id)
            else:
                self._nb_act[subs_id] += 1
                for sub_id_tmp in subs_id:
                    self._subs_impacted_to_act_id[sub_id_tmp].append(act_id)

            self._all_actions.append(copy.deepcopy(act))
            act_id += 1

        if len(self._all_actions) == 0:
            # no action provided... impossible to build myself
            raise RuntimeError("You provided only invalid actions or no action at all. Please check `list_act` "
                               "and provide a suitable set of actions.")

        if all_act_topo:
            # all provided actions affect topology
            # i need to check that some of them affect different substation
            if np.sum(subs_impact >= 1) <= 1:
                # all actions affect the same substation...
                raise RuntimeError("All your actions affect a substation, and it happens they all affect THE SAME "
                                   "substation. This is not possible to use RandomAct1 in such circumstances ("
                                   "mainly because we want to combine the action from the agent so...)")

    def _combine_actions(self, act1, act2):
        """some kind of "overload" of the `+` operator of grid2op to take into account the disconnected powerline"""
        res = act1 + act2
        for act in (act1, act2):
            set_status = act.line_set_status
            li = [(l_id, -1) for l_id, el in enumerate(set_status) if el == -1]
            if li:
                # force disconnection of the disconnected powerline in this action
                res.line_or_set_bus = li
                res.line_ex_set_bus = li
                res.line_set_status = li
        return res

    def sample_act(self, previous_act_id: Union[int, None] = None, sub_id_act: Union[int, None] = None):
        """
        This function has a simple behaviour: it samples, uniformly at random, one action among the possible actions.

        Possible actions are determined by `previous_act_id` and `sub_id_act` (both should be integers or None)

        `previous_act_id` is ignored if `sub_id_act` is not None: in this case, we don't want to sample an
        action acting on the substation `sub_id_act` so this will include `previous_act_id`

        """
        if previous_act_id is None and sub_id_act is None:
            act_id = self.space_prng.choice(len(self._all_actions))
        else:
            if sub_id_act is not None:
                # i cannot sample an action on this substation
                nb_act = np.sum(self._nb_act[:sub_id_act]) + np.sum(self._nb_act[(sub_id_act + 1):])
                act_id_tmp = self.space_prng.choice(nb_act)
                # now i need to convert it to a "real" id:
                # for that i need to find which "type" this action can be
                is_act_topo_before_sub_id = np.cumsum(self._nb_act[:sub_id_act])
                if act_id_tmp < is_act_topo_before_sub_id[-1]:
                    this_sub = np.where(act_id_tmp < is_act_topo_before_sub_id)[0][0]
                    if this_sub > 0:
                        act_id_tmp -= is_act_topo_before_sub_id[this_sub]
                    act_id = self._subs_impacted_to_act_id[this_sub][act_id_tmp]
                else:
                    act_id_tmp -= is_act_topo_before_sub_id[-1]
                    is_act_topo_before_sub_id = np.cumsum(self._nb_act[(sub_id_act + 1):])
                    this_sub = np.where(act_id_tmp < is_act_topo_before_sub_id)[0][0]
                    # this_sub is after sub_id_act, it's NOT the real substation ID !
                    real_sub_id = this_sub + (sub_id_act + 1)
                    if this_sub > 0:
                        act_id_tmp -= is_act_topo_before_sub_id[this_sub - 1]
                    act_id = self._subs_impacted_to_act_id[real_sub_id][act_id_tmp]
            else:
                # i sampled a previous action, but that does not come from a substation, i just need to
                # sample randomly from all actions except the previous one
                act_id = self.space_prng.choice(len(self._all_actions) - 1)
                if act_id >= previous_act_id:
                    act_id += 1

        this_random_act = self._all_actions[act_id]
        sub_id_act = self._subs_impacted_by_action[act_id]
        return act_id, sub_id_act, this_random_act

    def act(self, obs, reward, done=False):
        act_id, sub_id_act, this_random = self.sample_act()
        res = self._combine_actions(self.all_to_one, this_random)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import re
import os
import matplotlib.pyplot as plt
from leap_net.proxy.utils import create_env, reproducible_exp, DEFAULT_METRICS

from leap_net.ResNetLayer import ResNetLayer
from leap_net.agents import RandomN1, RandomN2
from leap_net.proxy.agentWithProxy import AgentWithProxy
from leap_net.proxy.proxyLeapNet import ProxyLeapNet
from leap_net.proxy.proxyBackend import ProxyBackend


def main(
        # reload the model
        env_name="l2rpn_case14_sandbox",
        save_path="model_saved",
        use_lightsim_if_available=True,
        val_regex=".*99[0-9].*",
        model_name="leapnet_case_14",
        layer=ResNetLayer,  # for now this is not serialized in the model
        # parameters for the evaluation
        do_dc=True,
        do_N1 = True,
        do_N2 = True,
        li_batch_size=tuple(),  # if you want to study the impact of the batch size
        total_evaluation_step=int(1024) * int(128),
        pred_batch_size=int(1024) * int(128),
        save_path_final_results="model_results",  # where the information about the prediction will be stored
        metrics=DEFAULT_METRICS,  # which metrics are used to evaluate the performance of the model
        verbose=1,  # do I print the results of the model
        # enforce reproducibility
        chron_id_val=0,
        env_seed=0,
        agent_seed=42,
):
    # create the environment
    env = create_env(env_name, use_lightsim_if_available=use_lightsim_if_available)

    env.chronics_handler.set_filter(lambda path: re.match(val_regex, path) is not None)
    env.chronics_handler.real_data.reset()
    obs = env.reset()

    if save_path_final_results is not None:
        if not os.path.exists(save_path_final_results):
            os.mkdir(save_path_final_results)
            if verbose > 0:
                print(f"Creating path \"{save_path_final_results}\" where results are stored")
    if do_dc:
        print("####################### \n"
              "## DC approximation  ## \n"
              "####################### \n")
        actor_evalN1_dc = RandomN1(env.action_space)
        proxy_dc = ProxyBackend(env._init_grid_path,
                                name=f"{model_name}_evalDC",
                                is_dc=True)
        proxy_dc.init([obs])  # dc is not stored (of course) so i need to manually load it
        agent_with_proxy_dc = AgentWithProxy(actor_evalN1_dc,
                                             proxy=proxy_dc,
                                             logdir=None)
        reproducible_exp(env,
                         agent=actor_evalN1_dc,
                         env_seed=env_seed,
                         agent_seed=agent_seed,
                         chron_id_start=chron_id_val)
        agent_with_proxy_dc.evaluate(env,
                                     load_path=None,
                                     save_path=save_path_final_results,
                                     # I do less because the current implementation takes too long
                                     total_evaluation_step=int(total_evaluation_step / 128),
                                     metrics=metrics,
                                     verbose=verbose
                                     )

    if do_N1:
        print("#######################\n"
              "##     Test set      ##\n"
              "#######################\n")
        actor_evalN1 = RandomN1(env.action_space)
        proxy_eval = ProxyLeapNet(name=f"{model_name}_evalN1",
                                  max_row_training_set=max(total_evaluation_step, pred_batch_size),
                                  eval_batch_size=pred_batch_size,
                                  layer=layer
                                  )
        agent_with_proxy_evalN1 = AgentWithProxy(actor_evalN1,
                                                 proxy=proxy_eval,
                                                 logdir=None)
        reproducible_exp(env,
                         agent=actor_evalN1,
                         env_seed=env_seed,
                         agent_seed=agent_seed,
                         chron_id_start=chron_id_val)
        agent_with_proxy_evalN1.evaluate(env,
                                         total_evaluation_step=total_evaluation_step,
                                         load_path=os.path.join(save_path, model_name),
                                         save_path=save_path_final_results,
                                         metrics=metrics,
                                         verbose=verbose
                                         )

    if do_N2:
        print("#######################\n"
              "##   SuperTest set   ##\n"
              "#######################\n")
        actor_evalN2 = RandomN2(env.action_space)
        proxy_eval = ProxyLeapNet(name=f"{model_name}_evalN2",
                                  max_row_training_set=max(total_evaluation_step, pred_batch_size),
                                  eval_batch_size=pred_batch_size,
                                  layer=layer
                                  )
        agent_with_proxy_evalN2 = AgentWithProxy(actor_evalN2,
                                                 proxy=proxy_eval,
                                                 logdir=None)
        reproducible_exp(env,
                         agent=actor_evalN2,
                         env_seed=env_seed,
                         agent_seed=agent_seed,
                         chron_id_start=chron_id_val)
        agent_with_proxy_evalN2.evaluate(env,
                                         total_evaluation_step=total_evaluation_step,
                                         load_path=os.path.join(save_path, model_name),
                                         save_path=save_path_final_results,
                                         metrics=metrics,
                                         verbose=verbose
                                         )

    if len(li_batch_size):
        print("###########################\n"
              "##  Impact of Batch size ##\n"
              "###########################\n")
        actor_batch_size = RandomN1(env.action_space)
        times_per_pf_ms = []
        total_times_ms = []
        for pred_batch_size in li_batch_size:
            reproducible_exp(env,
                             agent=actor_batch_size,
                             env_seed=env_seed,
                             agent_seed=agent_seed,
                             chron_id_start=chron_id_val)
            proxy_eval_tmp = ProxyLeapNet(name=f"{model_name}_evalN1_{pred_batch_size}",
                                          max_row_training_set=max(total_evaluation_step, pred_batch_size),
                                          eval_batch_size=pred_batch_size,  # min(total_evaluation_step, 1024*64)
                                          layer=layer)
            agent_with_tmp = AgentWithProxy(actor_batch_size,
                                            proxy=proxy_eval_tmp,
                                            logdir=None)

            dict_metrics = agent_with_tmp.evaluate(env,
                                                   total_evaluation_step=pred_batch_size,
                                                   load_path=os.path.join(save_path, model_name),
                                                   save_path=save_path_final_results,
                                                   metrics={},
                                                   verbose=0,
                                                   save_values=False  # I do not save the arrays
                                                   )
            total_pred_time_ms = 1000.*dict_metrics["predict_time"]
            total_times_ms.append(total_pred_time_ms)
            times_per_pf_ms.append(total_pred_time_ms/pred_batch_size)
            if verbose:
                print(f'Time to compute {pred_batch_size} powerflows: {total_pred_time_ms:.2f}ms '
                      f'({total_pred_time_ms/pred_batch_size:.4f} ms/powerflow)')

        if save_path_final_results is not None:
            # save the figure of the total time
            fig, ax = plt.subplots()
            ax.plot(li_batch_size, total_times_ms)
            ax.set_title('Total computation time')
            ax.set_xlabel("Number of powerflows")
            ax.set_ylabel("Total time (ms)")
            ax.set_yscale('log')
            ax.set_xscale('log')
            fig.tight_layout()
            fig.savefig(os.path.join(save_path_final_results, f"{model_name}_total_comp_time.pdf"))

            # save the figure of the total time
            fig, ax = plt.subplots()
            ax.plot(li_batch_size, times_per_pf_ms)
            ax.set_title('Computation time per powerflow')
            ax.set_xlabel("Number of powerflows")
            ax.set_ylabel("Time per powerflow (ms / pf)")
            ax.set_yscale('log')
            ax.set_xscale('log')
            fig.tight_layout()
            fig.savefig(os.path.join(save_path_final_results, f"{model_name}_time_per_pf.pdf"))


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import grid2op
from grid2op.Parameters import Parameters
import numpy as np
from leap_net.agents import RandomSub1, RandomSub2, RandomRefSub1

import unittest


class Test(unittest.TestCase):
    def setUp(self) -> None:
        # init the environment
        param = Parameters()
        param.NO_OVERFLOW_DISCONNECTION = True
        # i can act on all powerline / substation at once
        param.MAX_LINE_STATUS_CHANGED = 999999
        param.MAX_SUB_CHANGED = 999999
        # i can act every step on every line / substation (no cooldown)
        param.NB_TIMESTEP_COOLDOWN_LINE = 0
        param.NB_TIMESTEP_COOLDOWN_SUB = 0

        self.env = grid2op.make("l2rpn_case14_sandbox", param=param)
        self.obs = self.env.reset()

    def test_RandomSub1(self):
        agent1 = RandomSub1(self.env.action_space)
        agent1.seed(0)

        subs_index = []
        prev = 0
        for sub_id in range(self.obs.n_sub):
            nb_el = self.obs.sub_info[sub_id]
            subs_index.append((prev, prev + nb_el))
            prev += self.obs.sub_info[sub_id]

        for i in range(1000):
            act = agent1.act(None, None, None)
            lines_, subs_ = act.get_topological_impact()
            assert np.sum(lines_) == 0, f"error: RandomSub1 act on powerlines at iteration {i}"

            nb_sub_aff = 0
            for sub_id in range(self.env.n_sub):
                beg_, end_ = subs_index[sub_id]
                this_sub_topo = act.set_bus[beg_:end_]
                nb_sub_aff += np.any(this_sub_topo != 1)
            assert nb_sub_aff == 1, f"error: RandomSub1 act on multiple powerlines at iteration {i}"

    def test_RandomSub2(self):
        agent2 = RandomSub2(self.env.action_space)
        agent2.seed(0)

        subs_index = []
        prev = 0
        for sub_id in range(self.obs.n_sub):
            nb_el = self.obs.sub_info[sub_id]
            subs_index.append((prev, prev + nb_el))
            prev += self.obs.sub_info[sub_id]

        for i in range(1000):
            act = agent2.act(None, None, None)
            lines_, subs_ = act.get_topological_impact()
            assert np.sum(lines_) == 0

            nb_sub_aff = 0
            for sub_id in range(self.env.n_sub):
                beg_, end_ = subs_index[sub_id]
                this_sub_topo = act.set_bus[beg_:end_]
                nb_sub_aff += np.any(this_sub_topo != 1)
            assert nb_sub_aff == 2, f"error RandomSub2 does not act on 2 substation at iteration {i}"

    def test_RandomRefSub1(self):
        p = 0.1
        n = 1000
        agentref1 = RandomRefSub1(self.env.action_space, p=p)
        agentref1.seed(1)

        subs_index = []
        prev = 0
        for sub_id in range(self.obs.n_sub):
            nb_el = self.obs.sub_info[sub_id]
            subs_index.append((prev, prev + nb_el))
            prev += self.obs.sub_info[sub_id]

        test_ = [0, 0]
        for i in range(n):
            act = agentref1.act(None, None, None)
            lines_, subs_ = act.get_topological_impact()
            assert np.sum(lines_) == 0

            nb_sub_aff = 0
            for sub_id in range(self.env.n_sub):
                beg_, end_ = subs_index[sub_id]
                this_sub_topo = act.set_bus[beg_:end_]
                nb_sub_aff += np.any(this_sub_topo != 1)
            assert nb_sub_aff <= 1
            test_[nb_sub_aff] += 1
        # now check that i have approximately the right proba
        assert abs(test_[0] / n - (1 - p)) <= 1.96 * np.sqrt(p * (1-p)) / np.sqrt(n)
        assert abs(test_[1] / n - p) <= 1.96 * np.sqrt(p * (1-p)) / np.sqrt(n)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

from leap_net.Ltau import Ltau
from leap_net.ResNetLayer import ResNetLayer

__version__ = "0.0.4"
__all__ = ["Ltau", "ResNetLayer", "LtauNoAdd"]

try:
    from leap_net.generate_data import generate_dataset
    __all__ += ["generate_dataset"]
except ImportError:
    pass

try:
    from leap_net.kerasutils import MultipleDasetCallBacks
    __all__ += ["MultipleDasetCallBacks"]
except ImportError:
    pass

try:
    from leap_net.proxy import BaseProxy, BaseNNProxy, ProxyBackend, ProxyLeapNet, AgentWithProxy
    __all__ += ["BaseProxy", "BaseNNProxy", "ProxyBackend", "ProxyLeapNet", "AgentWithProxy"]
except ImportError:
    pass



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.
import os
import json
import re

import numpy as np
import tensorflow as tf
from tqdm import tqdm

from collections.abc import Iterable
from grid2op.Agent import BaseAgent

from leap_net.proxy.proxyLeapNet import ProxyLeapNet
import numpy as np

try:
    import matplotlib.pyplot as plt
    MATPLOT_OK = True
except ImportError:
    MATPLOT_OK = False

# TODO merge "reproducible exp" as a method of AgentWithProxy
# TODO implement a "I have gathered enough data, now let me learn without gathering more"


class PlotErrorOnGrid:
    """
    TODO move this class in this own file
    this class is used to "project" on the grid the metrics / errors of some proxies.

    it just ensure the projection, and as of creation (October, 26th 2020) it requires a development version
    of grid2op found at https://github.com/BDonnot/Grid2Op

    """
    def __init__(self, env):
        from grid2op.PlotGrid import PlotMatplot
        self.plot_helper = PlotMatplot(env.observation_space)
        self._line_attr = {"a_or", "a_ex", "p_or", "p_ex", "q_or", "q_ex", "v_or", "v_ex"}
        self._load_attr = {"load_p", "load_q", "load_v"}
        self._prod_attr = {"prod_p", "prod_q", "prod_v"}

    def get_fig(self, attr_nm, metrics):
        fig = None
        try:
            # only floating point values are supported at the moment
            metrics = metrics.astype(np.float)
        except Exception as exc_:
            return None

        if np.all(~np.isfinite(metrics)):
            # no need to plot a "all nan" vector
            return None

        if attr_nm in self._prod_attr:
            # deals generator attributes
            self.plot_helper.assign_gen_palette(increase_gen_size=1.5)
            fig = self.plot_helper.plot_info(gen_values=metrics, coloring="gen")
            self.plot_helper.restore_gen_palette()
        elif attr_nm in self._line_attr:
            # deals with lines attributes
            self.plot_helper.assign_line_palette()
            fig = self.plot_helper.plot_info(line_values=metrics, coloring="line")
            self.plot_helper.restore_line_palette()
        return fig


class AgentWithProxy(BaseAgent):
    """
    This class allows to easily manipulate an proxy (training et evaluating) coupled with "an actor" that allows
    for easily evaluate how a proxy deals with data coming from different distribution.

    We do not recommend to modify this class.
    """
    def __init__(self,
                 actor,  # the agent that will take some actions
                 proxy,  # the proxy to train / evaluate
                 logdir=None,  # tensorboard logs
                 update_tensorboard=256,  # tensorboard is updated every XXX training iterations
                 save_freq=int(1024)*int(64),  # model is saved every save_freq training iterations
                 ext=".h5",  # extension of the file in which you want to save the proxy
                 nb_obs_init=256,  # number of observations that are sent to the proxy to be initialized
                 ):
        BaseAgent.__init__(self, actor.action_space)
        self.actor = actor

        # to fill the training / test dataset
        self.global_iter = 0
        self.train_iter = 0
        self.__is_init = False  # is this model initiliazed
        self.is_training = True
        self._nb_obs_init = nb_obs_init

        # proxy part
        self._proxy = proxy

        # tensorboard (should be initialized after the proxy)
        if logdir is not None:
            logpath = os.path.join(logdir, self.get_name())
            self._tf_writer = tf.summary.create_file_writer(logpath, name=self.get_name())
        else:
            self._tf_writer = None
        self.update_tensorboard = update_tensorboard
        self.save_freq = int(save_freq)

        # save / load
        if re.match(r"^\.", ext) is None:
            # add a point at the beginning of the extension
            self.ext = f".{ext}"
        else:
            self.ext = ext
        self.save_path = None

    def init(self, env):
        """
        Initialize this object.

        It will first run the environment using the actor on a certain number of steps go gather some information
        and will then initialize the proxy wit these observations.

        Parameters
        ----------
        env:
            The grid2op environment
        """
        if self.__is_init:
            return
        # if the model is in "training mode" then when need to update the proxy with some observations of the
        # environment (basically to compute the scalers for the data)
        if self.is_training:
            # generate a few observation to init the proxy
            obss = []
            nb_obs = 0
            while nb_obs <= self._nb_obs_init:
                done = False
                reward = env.reward_range[0]
                obs = env.reset()
                while not done:
                    act = self.actor.act(obs, reward, done)
                    obs, reward, done, info = env.step(act)
                    if not done:
                        nb_obs += 1
                        obss.append(obs)

            # now build the poxy
            self._proxy.init(obss)

        # build the model
        self._proxy.build_model()
        self.__is_init = True

    # agent interface
    def act(self, obs, reward, done=False):
        """
        Overloading of the grid2op.BaseAgent "act" function that will first store the relevant attribute of the
        observation in the proxy.

        Then, if the proxy need to be trained, train it.

        And if the proxy need to be saved, save it.

        Parameters
        ----------
        obs: ``grid2op.Observation`
            The current grid2op observation.

        reward: ``float``
            The reward of the last action, irrelevant for this class but forwarded to the actor.

        done: ``bool``
            Whether or not there was a game over in the last action. Irrelevant for this class
             but forwarded to the actor.

        Returns
        -------
        act: `grid2op.Action`
            The action chosen by the actor (remember the proxy does not take any decisions here)
        """
        self.global_iter += 1
        self._store_obs(obs)
        if self.is_training:
            batch_losses = self._proxy.train(tf_writer=self._tf_writer)
            if batch_losses is not None:
                self.train_iter += 1
                self._save_tensorboard(batch_losses)
                self._save_model()
        return self.actor.act(obs, reward, done)

    def train(self, env, total_training_step, save_path=None, load_path=None, verbose=1):
        """
        Completely train the proxy. This run a possible entire procedure to train a proxy.

        We highly recommend to call this method to train it.

        Parameters
        ----------
        env:
            The environment

        total_training_step:
            The name is completely misleading. This is not the total number of training step, but the total
            number of data that will be see during training. For example, if the batch size is 32, then
            then `"total number of training step" = "total number of data that will be see during training" / 32`
            this number is "total number of data that will be see during training".

        save_path: ``str``
            Path where the proxy and some other data from this instance will be saved. (``None`` to deactivate it)

        load_path: ``str``
            If it is not None, data stored at the location "path" will be loaded back.

        verbose: ``int``
            Degree of verbosity. The more verbose the more information will be plotted on the command line

        """
        done = False
        reward = env.reward_range[0]
        self.save_path = save_path
        if self.save_path is not None:
            if not os.path.exists(self.save_path):
                os.mkdir(self.save_path)
        if load_path is not None:
            self.load(load_path)
        self.is_training = True
        if not self.__is_init:
            self.init(env)
        with tqdm(total=total_training_step, disable=verbose == 0) as pbar:
            # update the progress bar
            pbar.update(self.global_iter)
            obs = self._reboot(env)
            # and do the "gym loop"
            while not done:
                act = self.act(obs, reward, done)
                # TODO handle multienv here
                if not self._proxy.DEBUG or self.global_iter <= self._proxy.train_batch_size:
                    obs, reward, done, info = env.step(act)
                    if done:
                        obs = self._reboot(env)
                        done = False
                # TODO have  multiprocess here: this script stores the data in the database of the proxy
                # TODO the proxy reads from this database (exposed as a tensorflow dataset) to train the model
                pbar.update(1)
                if self.global_iter >= total_training_step:
                    break

        # save the model at the end
        self.save(self.save_path)

    def evaluate(self, env, total_evaluation_step, load_path, save_path=None, metrics=None,
                 verbose=0, save_values=True):
        """
        This is a function to evaluate the performance of a proxy.

        We highly recommend to use this method if you coded a model that follows the "proxy" interface.

        Parameters
        ----------
        env:
            The grid2op environment used

        total_evaluation_step:
            Total number of states the proxy will be evaluated on

        load_path:
            Path from which the proxy will be loaded.

        metrics:
            dictionary of function, with keys being the metrics name, and values the function that compute
            this metric (on the whole output) that should be `metric_fun(y_true, y_pred)`

        save_path: ``str``
            Path where the results of the models are stored. This is not the same as the "save_path" argument
            of "train")

        verbose: ``int``
            Degrees of verbosity (the higher the more information will be printed on the command prompt)

        save_values: ``bool``
            Do you save the computed values (prediction of the proxy AND ground truth) or just the metrics

        Returns
        -------
        res: ``dict``
            The dictionary containing the values of each metrics defined in "metrics" for each output variable
            of the proxy.

        """
        if not os.path.exists(save_path):
            print(f"Creating path \"{save_path}\" to save the logs of the trained agent")
            os.mkdir(save_path)

        error_plot = None
        try:
            error_plot = PlotErrorOnGrid(env)
        except Exception as exc_:
            # plotting will not be available, but this is not a reason to crash
            pass

        done = False
        reward = env.reward_range[0]
        self.is_training = False
        self.save_path = None  # disable the saving of the model

        self.load(load_path)
        self.global_iter = 0
        self.train_iter = 0

        # TODO find a better approach for more general proxy that can adapt to grid of different size
        sizes = self._proxy.get_output_sizes()
        true_val = [np.zeros((total_evaluation_step, el), dtype=self._proxy.dtype) for el in sizes]
        pred_val = [np.zeros((total_evaluation_step, el), dtype=self._proxy.dtype) for el in sizes]

        if not self.__is_init:
            self.init(env)
        with tqdm(total=total_evaluation_step, disable=verbose == 0) as pbar:
            # update the progress bar
            pbar.update(self.global_iter)
            obs = self._reboot(env)
            # and do the "gym loop"
            while not done:
                # act (which increment global_iter and ask the actor what action to do)
                act = self.act(obs, reward, done)

                # save the predictions and the reference
                predictions = self._proxy.predict(force=self.global_iter == total_evaluation_step)
                if predictions is not None:
                    for arr_, pred_ in zip(pred_val, predictions):
                        sz = pred_.shape[0]
                        min_ = max((self.global_iter-sz), 0)
                        arr_[min_:self.global_iter, :] = pred_

                reality = self._proxy.get_true_output(obs)
                for arr_, ref_ in zip(true_val, reality):
                    arr_[self.global_iter-1, :] = ref_.reshape(-1)

                # TODO handle multienv here (this might be more complicated!)
                obs, reward, done, info = env.step(act)
                if done:
                    obs = self._reboot(env)
                    done = False
                pbar.update(1)
                if self.global_iter >= total_evaluation_step:
                    break
        # save the results and compute the metrics
        # TODO save the real x's too!
        return self._save_results(obs, save_path, metrics, pred_val, true_val, verbose, save_values, error_plot)

    def save(self, path):
        """
        Part of the l2rpn_baselines interface, this allows to save a model. Its name is used at saving time. The
        same name must be reused when loading it back.

        Parameters
        ----------
        path: ``str``
            The path where to save the agent.

        """
        if path is not None:
            path_save = os.path.join(path, self.get_name())
            if not os.path.exists(path_save):
                os.mkdir(path_save)
            self._save_metadata(path_save)
            self._proxy.save_data(path=path_save, ext=self.ext)

    def load(self, path):
        """
        Load a previously stored experiments from the hard drive.

        This both load data for this class and from the proxy.

        Parameters
        ----------
        path: ``str``
            Where to load the experiment from.

        """
        if path is not None:
            # the following if is to be able to restore a file with possibly a different name...
            if self.is_training:
                path_model = self._get_path_nn(path, self.get_name())
            else:
                path_model = path
            if not os.path.exists(path_model):
                raise RuntimeError(f"You asked to load a model at \"{path_model}\" but there is nothing there.")
            self._load_metadata(path_model)
            self._proxy.build_model()
            self._proxy.load_data(path=path, ext=self.ext)

    def get_name(self):
        """get the name of this experiment, that by definition (for now) is the name given to the proxy"""
        return self._proxy.name

    ###############
    ## Utilities ##
    ##############
    def _store_obs(self, obs):
        """
        Utility function that only (for now) tell the proxy to store the observation.

        Parameters
        ----------
        obs: `grid2op.Action.BaseObservation`
            The current observation

        """
        self._proxy.store_obs(obs)

    # save load model
    def _get_path_nn(self, path, name):
        """utilities when path and file names are not formatted the same way"""
        if name is None:
            path_model = path
        else:
            path_model = os.path.join(path, name)
        return path_model

    def _save_metadata(self, path_model):
        """save the dimensions of the models and the scalers"""
        json_nm = "metadata.json"
        me = self._to_dict()
        me["proxy"] = self._proxy.get_metadata()
        with open(os.path.join(path_model, json_nm), "w", encoding="utf-8") as f:
            json.dump(obj=me, fp=f)

    def _load_metadata(self, path_model):
        """load the metadata of the experiments (both for me and for the proxy)"""
        json_nm = "metadata.json"
        with open(os.path.join(path_model, json_nm), "r", encoding="utf-8") as f:
            me = json.load(f)
        self._from_dict(me)
        self._proxy.load_metadata(me["proxy"])

    def _to_dict(self):
        """output a json serializable dictionary representing the current state of the experiment"""
        res = {}
        res["train_iter"] = int(self.train_iter)
        res["global_iter"] = int(self.global_iter)
        return res

    def _from_dict(self, dict_):
        """update my meta data only, do not affect the proxy"""
        self.train_iter = int(dict_["train_iter"])
        self.global_iter = int(dict_["global_iter"])

    def _save_tensorboard(self, batch_losses):
        """save all the information needed in tensorboard."""
        if self._tf_writer is None:
            return

        # Log some useful metrics every even updates
        if self.train_iter % self.update_tensorboard == 0:
            with self._tf_writer.as_default():
                # save total loss
                tf.summary.scalar(f"0_global_loss",
                                  batch_losses[0],
                                  self.train_iter,
                                  description="Loss of the entire model")
                self._proxy.save_tensorboard(self._tf_writer, self.train_iter, batch_losses[1:])

    def _save_model(self):
        """trigger the saving of the model"""
        if self.train_iter % self.save_freq == 0:
            self.save(self.save_path)

    def _save_results(self, obs, save_path, metrics, pred_val, true_val, verbose,
                      save_values=True, error_plot=None):
        """
        This function will save the results of the evaluation of the proxy in multiple form:

        - all the "ground truth" for all the variable that the proxy needs to predict
        - all the predicted values for all the variables that the proxy predicted
        - the timing and all metrics in a "metrics.json" file
        - the metadata of the model in the "metada.json" file
        - the data of the model (depending on the model)
        - for all metrics that have been computed, an attempt to display on the figure where the errors are located (
          it uses the grid2op.PlotGrid module)

        Parameters
        ----------
        obs
        save_path
        metrics
        pred_val
        true_val
        verbose
        save_values: ``bool``
            Do I save the arrays (of true and predicted values). If ``False`` only the json is saved
        error_plot:
            Utility to plot the error on the grid in a matplotlib figures

        Returns
        -------

        """
        # TODO save the "x" given to the proxy here too

        # compute the metrics (if any)
        dict_metrics = {}
        dict_metrics["predict_step"] = int(self.global_iter)
        dict_metrics["predict_time"] = float(self._proxy.get_total_predict_time())
        dict_metrics["avg_pred_time_s"] = float(self._proxy.get_total_predict_time()) / float(self.global_iter)

        if metrics is not None:
            array_names = self._proxy.get_attr_output_name(obs)
            for metric_name, metric_fun in metrics.items():
                dict_metrics[metric_name] = {}
                for nm, pred_, true_ in zip(array_names, pred_val, true_val):
                    tmp = metric_fun(true_, pred_)
                    # print the results and make sure the things are json serializable
                    if isinstance(tmp, Iterable):
                        if verbose >= 2:
                            print(f"{metric_name} for {nm}: {tmp}")  # don't print not to overload the display
                        dict_metrics[metric_name][nm] = [float(el) for el in tmp]

                        # plot the error on the grid layout
                        if error_plot is not None and save_path is not None and MATPLOT_OK:
                            fig = error_plot.get_fig(nm, tmp)
                            if fig is not None:
                                save_path_fig = os.path.join(save_path, self.get_name())
                                if not os.path.exists(save_path_fig):
                                    os.mkdir(save_path_fig)
                                fig.savefig(os.path.join(save_path_fig, f"{metric_name}_{nm}.pdf"))
                                plt.close(fig)
                    else:
                        if verbose >=1:
                            print(f"{metric_name} for {nm}: {tmp:.2f}")
                        dict_metrics[metric_name][nm] = float(tmp)

        # save the numpy arrays (if needed)
        if save_path is not None:
            # save the proxy and the meta data
            self.save(save_path)

            # now the other data
            array_names = self._proxy.get_attr_output_name(obs)
            save_path = os.path.join(save_path, self.get_name())
            if not os.path.exists(save_path):
                os.mkdir(save_path)

            if save_values:
                for nm, pred_, true_ in zip(array_names, pred_val, true_val):
                    np.save(os.path.join(save_path, f"{nm}_pred.npy"), pred_)
                    np.save(os.path.join(save_path, f"{nm}_real.npy"), true_)
            with open(os.path.join(save_path, "metrics.json"), "w", encoding="utf-8") as f:
                json.dump(dict_metrics, fp=f, indent=4, sort_keys=True)
        return dict_metrics

    def _reboot(self, env):
        """when an environment is "done" this function reset it and act a first time with the agent_action"""
        # TODO skip and random start at some steps
        done = False
        reward = env.reward_range[0]
        obs = env.reset()
        obs, reward, done, info = env.step(self.actor.act(obs, reward, done))
        while done:
            # we restart until we find an environment that is not "game over"
            obs = env.reset()
            obs, reward, done, info = env.step(self.actor.act(obs, reward, done))
        return obs



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of leap_net, leap_net a keras implementation of the LEAP Net model.

import re
import os

import tensorflow as tf

from leap_net.ResNetLayer import ResNetLayer
from leap_net.agents import RandomNN1
from leap_net.proxy.agentWithProxy import AgentWithProxy
from leap_net.proxy.proxyLeapNet import ProxyLeapNet

from leap_net.proxy.utils import create_env, reproducible_exp, DEFAULT_METRICS, limit_gpu_usage
# TODO resuse the "main" from train_case14 (only the selection of data change between the 2)


def main(limit_gpu_memory=True,
         total_train=int(1024)*int(1024),  # number of observations that will be gathered
         env_name="l2rpn_neurips_2020_track2_small",
         # log during training
         save_path="model_saved_118",
         save_path_tensorbaord="tf_logs",
         update_tensorboard=256,  # tensorboard is updated every XXX training iterations
         save_freq=int(1024) * int(4),  # model is saved every save_freq training iterations
         ext=".h5",  # extension of the file in which you want to save the proxy
         nb_obs_init=512,  # number of observations that are sent to the proxy to be initialized
         # dataset / data generation part
         env_seed=42,
         agent_seed=1,
         use_lightsim_if_available=True,
         val_regex_train=".*Scenario_february_[1-9][0-9].*",  # be careful to remember that for the validation set !
         actor_class=RandomNN1,
         # perform an evaluation on the training set at the end of training
         eval_training_set=int(1024) * int(128),  # number of powerflow the proxy will do after training
         pred_batch_size=int(1024) * int(32),  # number of powerflow that will be done by the proxy "at once"
         save_path_final_results="model_results",  # where the information about the prediction will be stored
         metrics=DEFAULT_METRICS,  # which metrics are used to evaluate the performance of the model
         verbose=1,  # do I print the results of the model
         # proxy part (sizes are not really "parametrized" this is just "something that works approximately)
         model_name="leapnet_case_118",
         sizes_enc=(60, 60, 60,),
         sizes_main=(300, 300, 300, 300, 300),
         sizes_out=(60, ),
         lr=1e-4,
         layer=ResNetLayer,
         layer_act=None,
         scale_main_layer=600,
         scale_input_dec_layer=None,
         scale_input_enc_layer=None,
         attr_x=("prod_p", "prod_v", "load_p", "load_q"),
         attr_y=("a_or", "a_ex", "p_or", "p_ex", "q_or", "q_ex", "prod_q", "load_v", "v_or", "v_ex"),
         attr_tau=("line_status",),
         # TODO add the other constructor parameters of the proxy
         ):
    if limit_gpu_memory:
        limit_gpu_usage()

    # create the environment
    env = create_env(env_name, use_lightsim_if_available=use_lightsim_if_available)

    # I select only part of the data, for training
    env.chronics_handler.set_filter(lambda path: re.match(val_regex_train, path) is not None)
    env.chronics_handler.real_data.reset()
    obs = env.reset()

    # now train the agent
    actor = actor_class(env.action_space, p=0.5)
    reproducible_exp(env, actor, env_seed=env_seed, agent_seed=agent_seed)

    proxy = ProxyLeapNet(name=model_name,
                         lr=lr,
                         layer=layer,
                         layer_act=layer_act,
                         sizes_enc=sizes_enc,
                         sizes_main=sizes_main,
                         sizes_out=sizes_out,
                         scale_main_layer=scale_main_layer,
                         scale_input_dec_layer=scale_input_dec_layer,
                         scale_input_enc_layer=scale_input_enc_layer,
                         attr_x=attr_x,
                         attr_tau=attr_tau,
                         attr_y=attr_y)
    agent_with_proxy = AgentWithProxy(actor,
                                      proxy=proxy,
                                      logdir=save_path_tensorbaord,
                                      update_tensorboard=update_tensorboard,
                                      save_freq=save_freq,
                                      ext=ext,
                                      nb_obs_init=nb_obs_init
                                      )

    # train it
    agent_with_proxy.train(env,
                           total_train,
                           save_path=save_path
                           )

    if verbose:
        print("Summary of the model used:")
        print(proxy._model.summary())

    if eval_training_set is not None:
        print("Evaluation of the model on the training set")
        proxy_eval = ProxyLeapNet(name=f"{model_name}_evalTrainSet",
                                  max_row_training_set=max(eval_training_set, pred_batch_size),
                                  eval_batch_size=pred_batch_size,  # min(total_evaluation_step, 1024*64)
                                  layer=layer)
        agent_with_proxy_eval = AgentWithProxy(actor,
                                               proxy=proxy_eval,
                                               logdir=None)
        reproducible_exp(env, actor, env_seed=env_seed, agent_seed=agent_seed)
        agent_with_proxy_eval.evaluate(env,
                                       total_evaluation_step=pred_batch_size,
                                       load_path=os.path.join(save_path, model_name),
                                       save_path=save_path_final_results,
                                       metrics=metrics,
                                       verbose=verbose
                                       )


if __name__ == "__main__":
    main()
